"""This module is the single source for this package's version."""

import importlib.metadata
import re


def get_version() -> str:
    """Get package version.

    If this is a pre-release, reformat to be compatible with semver.

    Returns:
        Version string.
    """
    version = importlib.metadata.version(__package__)
    version_regex = re.compile(r"([0-9]+)\.([0-9]+)\.([0-9]+)")
    match = re.match(version_regex, version).group()

    if match:
        suffix = version.split(match)[-1]
        if suffix and not suffix.startswith("-"):
            version = f"{match}-{suffix}"

    return version



"""This module is a script for normalizing the audio to have a peak amplitude
of 0.0 decibels.
"""

# pylint: disable=import-error
from argparse import ArgumentParser
from collections import defaultdict
from concurrent.futures import as_completed, ThreadPoolExecutor
import json
from pathlib import Path
from typing import Dict

import matplotlib.pyplot as plt
from pydub import AudioSegment, effects, utils
from tqdm import tqdm

from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.rekordbox_track import RekordboxTrack


def thread(track: RekordboxTrack, data_dict: Dict):
    """Normalize the amplitude of tracks to have 0.0 headroom.

    Overwrite the track's audio file with the normalized audio. Tags are
    retained and audio files are exported at 320 kbps. Collect the dB
    frequencies and filenames for analysis.

    Args:
        track: Track object.
        data_dict: Dictionary mapping amplitude to a frequency count and list
            of tracks.
    """
    loc = track.get_location()
    audio = AudioSegment.from_file(loc)
    if audio.max_dBFS < 0.0:
        audio = effects.normalize(audio, headroom=0.0)
        audio.export(
            loc,
            tags=utils.mediainfo(loc).get("TAG", {}),
            bitrate="320k",
        )
    # Round off these values now or else aggregate keys in the dB_data.json
    # file written at the end of this script.
    # NOTE: your bar chart will probably be hard to read unless you aggregate
    # these keys.
    data_dict[audio.max_dBFS]["count"] += 1
    data_dict[audio.max_dBFS]["tracks"].append(loc.name)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--collection", help="Path to collection")
    args = parser.parse_args()

    # Submit normalization jobs to a thread pool and collect dB data.
    collection = RekordboxCollection(Path(args.collection))
    tracks = collection.get_tracks().values()
    data = defaultdict(lambda: {"count": 0, "tracks": []})
    with tqdm(total=len(tracks)) as pbar:
        with ThreadPoolExecutor(max_workers=32) as pool:
            futures = [pool.submit(thread, track, data) for track in tracks]
            for future in as_completed(futures):
                future.result()
                pbar.update()

    # Write collected data to a JSON file.
    with open("dB_data.json", mode="w", encoding="UTF-8") as _file:
        json.dump(data, _file)

    # Plot log scale amplitudes.
    data = {k: data[k] for k in sorted(data)}
    plt.figure(figsize=(10, 6))
    plt.bar(list(data.keys()), [v["count"] for v in data.values()])
    plt.title("Collection Peak Amplitudes (log scale)", fontweight="bold")
    plt.xlabel("Amplitude", fontweight="bold")
    plt.ylabel("Frequency", fontweight="bold")
    plt.yscale("log")
    plt.tick_params(left=False, bottom=False)
    plt.tight_layout()
    plt.savefig("Collection Peak Amplitudes")



"""Testing for the config module."""

import os
import re

import pytest

from djtools.utils.config import UtilsConfig


def test_utilsconfig_aws_profile_not_set():
    """Test for the UtilsConfig class."""
    cfg = {"check_tracks": True}
    os.environ["AWS_PROFILE"] = ""  # pylint: disable=no-member
    with pytest.raises(
        RuntimeError,
        match=re.escape(
            "Without aws_profile set to a valid profile ('default' or otherwise) "
            "you cannot use the check_tracks feature"
        ),
    ):
        UtilsConfig(**cfg)


def test_utilsconfig_spotify_creds_warning(caplog):
    """Test the UtilsConfig class."""
    caplog.set_level("WARNING")
    cfg = {
        "check_tracks": True,
        "check_tracks_spotify_playlists": ["playlist"],
    }
    os.environ["AWS_PROFILE"] = "default"  # pylint: disable=no-member
    UtilsConfig(**cfg)
    assert caplog.records[0].message == (
        "check_tracks depends on valid Spotify API credentials in "
        "SpotifyConfig."
    )


def test_utilsconfig_recording_file_not_set():
    """Test for the UtilsConfig class."""
    cfg = {
        "process_recording": True,
        "recording_file": "notreal",
    }
    with pytest.raises(
        RuntimeError, match='Could not find recording_file "notreal"'
    ):
        UtilsConfig(**cfg)


def test_utilsconfig_recording_playlist_not_set():
    """Test for the UtilsConfig class."""
    cfg = {
        "process_recording": True,
        "recording_file": "",
        "recording_playlist": "",
    }
    with pytest.raises(
        RuntimeError,
        match=(
            "You must provide a playlist name as recording_playlist "
            "and this name must exists in spotify_playlists.yaml."
        ),
    ):
        UtilsConfig(**cfg)


@pytest.mark.parametrize("bit_rate", ["35", "321"])
def test_utilsconfig_validates_bitrate_error(bit_rate):
    """Test for the UtilsConfig class."""
    cfg = {"audio_bitrate": bit_rate}
    with pytest.raises(
        ValueError,
        match=re.escape("audio_bitrate must be in the range [36, 320]"),
    ):
        UtilsConfig(**cfg)


@pytest.mark.parametrize("bit_rate", ["36", "178", "320"])
def test_utilsconfig_validates_bitrate_success(bit_rate):
    """Test for the UtilsConfig class."""
    cfg = {"audio_bitrate": bit_rate}
    config = UtilsConfig(**cfg)
    assert config.audio_bitrate == str(bit_rate)


def test_utilsconfig_validate_format_warning(caplog):
    """Test for the UtilsConfig class."""
    caplog.set_level("WARNING")
    cfg = {"audio_format": "mp3", "normalize_audio": True}
    _ = UtilsConfig(**cfg)
    assert caplog.records[0].message == (
        "You must install FFmpeg in order to use non-wav file formats."
    )



"""This module contains fixtures for DJ Tools."""

import os
from argparse import Namespace
from pathlib import Path
from unittest import mock

import pytest
import yaml
from bs4 import BeautifulSoup
from pydub import AudioSegment, generators

from djtools.configs.config import BaseConfig
from djtools.collection.config import PlaylistConfig
from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.rekordbox_track import RekordboxTrack


@pytest.fixture
def namespace():
    """Test Namespace object fixture."""
    return Namespace(link_configs=False, log_level="INFO", version=False)


@pytest.fixture
@mock.patch("djtools.spotify.helpers.Client", mock.MagicMock())
def config():
    """Test config fixture."""
    return BaseConfig()


@pytest.fixture
def config_file_teardown():
    """Teardown config.yaml."""
    yield
    config_dir = Path(__file__).parent.parent / "src" / "djtools" / "configs"
    config_file = config_dir / "config.yaml"
    if config_file.exists():
        config_file.unlink()


@pytest.fixture(scope="session")
def input_tmpdir(tmpdir_factory):
    """Test tmpdir fixture."""
    return tmpdir_factory.mktemp("input")


@pytest.fixture(scope="session")
def audio_file(input_tmpdir):  # pylint: disable=redefined-outer-name
    """Test audio file fixture."""
    headroom = -2
    audio = AudioSegment.silent(duration=1000).overlay(
        generators.WhiteNoise()
        .to_audio_segment(duration=1000)
        .apply_gain(headroom)
    )
    _file = Path(input_tmpdir) / "file.wav"
    audio.export(_file, format="wav")

    return audio, _file


@pytest.fixture
def playlist_config():
    """Test playlist config fixture."""
    with open(
        "tests/data/collection_playlists.yaml", mode="r", encoding="utf-8"
    ) as _file:
        return yaml.load(_file.read(), Loader=yaml.FullLoader)


@pytest.fixture
def playlist_config_content():
    """Test playlist config content fixture."""
    with open(
        "tests/data/collection_playlists.yaml", mode="r", encoding="utf-8"
    ) as _file:
        return _file.read()


@pytest.fixture
def playlist_config_obj(
    playlist_config,
):  # pylint: disable=redefined-outer-name
    """Test playlist config object fixture."""
    return PlaylistConfig(**playlist_config)


@pytest.fixture(scope="session")
def rekordbox_playlist_tag():  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox playlist tag."""
    playlist_string = (
        """<NODE Name="ROOT" Type="0" Count="2">\n"""
        """  <NODE Name="Genres" Type="0" Count="1">\n"""
        """    <NODE Name="Hip Hop" Type="1" Entries="1">\n"""
        """      <TRACK Key="2"/>\n"""
        """    </NODE>\n"""
        """  </NODE>\n"""
        """  <NODE Name="My Tags" Type="0" Count="1">\n"""
        """    <NODE Name="Dark" Type="1" Entries="0"/>\n"""
        """  </NODE>\n"""
        """</NODE>"""
    )
    return BeautifulSoup(playlist_string, "xml").find("NODE")


@pytest.fixture(scope="session")
def rekordbox_track_tag(input_tmpdir):  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox track tag."""
    track_string = (
        """<TRACK Artist="A Tribe Called Quest" AverageBpm="86.00" """
        """Comments=" /* Gangsta */ " DateAdded="2022-06-24" Genre="Hip Hop / R&amp;B" """
        """Label="Label" Location="file://localhost/track2.mp3" """
        """Tonality="7B" Rating="0" TrackID="2" """
        """TrackNumber="2" Year="2022"/>"""
    )
    track_tag = BeautifulSoup(track_string, "xml").find("TRACK")
    test_dir = Path(input_tmpdir) / "input"
    test_dir.mkdir(exist_ok=True)
    track_name = Path(track_tag["Location"]).name
    prefix = (
        "file://localhost"
        if os.name == "posix"  # pylint: disable=no-member
        else "file://localhost/"
    )
    track_tag["Location"] = f"{prefix}{(test_dir / track_name).as_posix()}"
    with open(test_dir / track_name, mode="w", encoding="utf-8") as _file:
        _file.write("")

    return track_tag


@pytest.fixture()
def rekordbox_track(
    rekordbox_track_tag,
):  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox track object."""
    return RekordboxTrack(rekordbox_track_tag)


@pytest.fixture(scope="session")
def rekordbox_xml(input_tmpdir):  # pylint: disable=redefined-outer-name
    """Fixture for XML file."""
    input_tmpdir = Path(input_tmpdir)
    collection = RekordboxCollection("tests/data/rekordbox.xml")
    for track in collection.get_tracks().values():
        track.set_location(input_tmpdir / track.get_location().name)
        with open(track.get_location(), mode="w", encoding="utf-8") as _file:
            _file.write("")

    return collection.serialize(path=input_tmpdir / "rekordbox.xml")


@pytest.fixture(scope="session")
def rekordbox_collection_tag(
    rekordbox_xml,
):  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox collection tag."""
    with open(rekordbox_xml, mode="r", encoding="utf-8") as _file:
        xml = BeautifulSoup(
            _file.read(), "xml"
        )  # pylint: disable=redefined-outer-name

    return xml


@pytest.fixture(scope="session")
def rekordbox_collection(
    rekordbox_xml,
):  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox collection object."""
    return RekordboxCollection(rekordbox_xml)


@pytest.fixture(scope="session")
def rekordbox_playlist(
    rekordbox_collection,
):  # pylint: disable=redefined-outer-name
    """Fixture for Rekordbox playlist object."""
    return rekordbox_collection.get_playlists()


###############################################################################
# Pytest hooks for producing timing information for fixtures and test cases.
###############################################################################


# import time

# @pytest.hookimpl(hookwrapper=True)
# def pytest_fixture_setup(request):
#     start = time.time()
#     yield
#     print(f"[DEBUG] {request.fixturename} [fixture]={time.time() - start}")


# @pytest.hookimpl(hookwrapper=True)
# def pytest_runtest_setup(item):
#     start = time.time()
#     yield
#     print(f"[DEBUG] {item.listnames()[-1]} [setup]={time.time() - start}")


# @pytest.hookimpl(hookwrapper=True)
# def pytest_runtest_call(item):
#     start = time.time()
#     yield
#     print(f"[DEBUG] {item.listnames()[-1]} [call]={time.time() - start}")


# @pytest.hookimpl(hookwrapper=True)
# def pytest_runtest_teardown(item):
#     start = time.time()
#     yield
#     print(f"[DEBUG] {item.listnames()[-1]} [teardown]={time.time() - start}")



"""This script is used to print the time, in seconds, of hot cues in a
collection.
"""

# pylint: disable=consider-using-f-string,invalid-name,missing-class-docstring,missing-function-docstring,pointless-statement,redefined-outer-name
import json
import logging
import re

from argparse import ArgumentParser, Namespace
from pathlib import Path
from string import ascii_uppercase as uppercase_letters
from typing import List, Optional

logger = logging.getLogger(__name__)

try:
    from bs4 import BeautifulSoup, FeatureNotFound, Tag
except ModuleNotFoundError as exc:
    msg = (
        'The "BeautifulSoup" library is required; install with '
        "'pip install bs4'"
    )
    logger.critical(msg)
    raise RuntimeError(msg) from exc


def main() -> None:
    args = parse_args()

    setup_logging(args.log_level)

    all_tracks = get_tracks_from_collection(args.path)

    if args.tracks:
        tracks = find_tracks(all_tracks, args.tracks, args.strict_filtering)
    else:
        tracks = all_tracks

    for track in tracks:
        track.hot_cues

    if args.save:
        save_tracks(all_tracks, args.save_path)


class Track:
    def __init__(self, positions: List[Tag], name: str, artist: str) -> None:
        self._name = name
        self._artist = artist
        self._hot_cues = {
            uppercase_letters[index]: hot_cue
            for index, hot_cue in enumerate(
                sorted(positions, key=lambda x: int(x.get("Num", 0)))
            )
        }

    @classmethod
    def create(cls, tag: Tag) -> Optional["Track"]:
        name = tag.get("Name")
        artist = tag.get("Artist")

        if not name:
            logger.debug('Tag "%s" has no name' % str(tag))
            name = "NO NAME"

        if not artist:
            logger.debug('Tag "%s" has no artist' % str(tag))
            artist = "NO ARTIST"

        positions = tag.find_all("POSITION_MARK")
        if not positions:
            logger.warning('Track "%s - %s" has no hot cues!' % (name, artist))
            return None

        return cls(positions, name, artist)

    @property
    def name(self) -> str:
        return self._name

    @property
    def artist(self) -> str:
        return self._artist

    @property
    def hot_cues(self) -> None:
        logger.info('Hot cues for "%s" by "%s"' % (self.name, self.artist))
        for letter, hot_cue in self._hot_cues.items():
            logger.info("%s: %s" % (letter, hot_cue.get("Start")))

    def to_dict(self) -> str:
        return {
            "name": self.name,
            "artist": self.artist,
            "hot_cues": {
                letter: hot_cue.get("Start")
                for letter, hot_cue in self._hot_cues.items()
            },
        }


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("path", help="Path to a Rekordbox library XML export.")
    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Logging verbosity.",
    )
    parser.add_argument(
        "--tracks",
        action="append",
        type=str,
        help=(
            "One or more tracks whose hot cue timings will be displayed. "
            "These can be just the names of the tracks but, if "
            "'--strict-filtering' is enabled, they must follow this format:"
            "\n\t NAME - ARTIST."
        ),
    )
    parser.add_argument(
        "--strict-filtering",
        action="store_true",
        help=(
            "If enabled, tracks whose hot cues are displayed will only do so "
            "if both their name and artist match."
        ),
    )
    parser.add_argument(
        "--save",
        action="store_true",
        help="Whether or not to save all the hot cue data as a JSON file.",
    )
    parser.add_argument(
        "--save-path",
        default="track_hot_cues.json",
        help="Default path of the JSON file created when saving.",
    )
    args = parser.parse_args()

    return args


def setup_logging(log_level: str):
    logger.setLevel(log_level)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    formatter = logging.Formatter("%(levelname)s - %(message)s")
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)


def get_tracks_from_collection(xml_path: str) -> List[Track]:
    path = Path(xml_path)
    if not path.exists():
        msg = "Could not find %s; ensure it exists!" % xml_path
        logger.critical(msg)
        raise RuntimeError(msg)

    try:
        with open(path, mode="r", encoding="utf-8") as _file:
            doc = BeautifulSoup(_file.read(), "xml")
    except FeatureNotFound as exc:
        msg = (
            'Could not open %s; please install "lxml" by running '
            "'pip install lxml'" % str(path)
        )
        logger.critical(msg)
        raise RuntimeError(msg) from exc

    tracks = list(
        filter(
            None,
            [
                Track.create(t)
                for t in doc.find_all("TRACK")
                if t.get("Location")
            ],
        )
    )

    return tracks


def find_tracks(
    all_tracks: List[Track], tracks_to_find: List[str], strict_filtering: bool
) -> List[Track]:
    set_of_tracks_to_find = set(tracks_to_find)
    matching_tracks = []

    for track in all_tracks:
        if strict_filtering:
            track_id = "%s - %s" % (track.name, track.artist)
            if track_id not in set_of_tracks_to_find:
                continue
            matching_tracks.append(track)
            continue

        if any(
            re.match(track.name, track_to_find)
            for track_to_find in set_of_tracks_to_find
        ):
            matching_tracks.append(track)

    return matching_tracks


def save_tracks(tracks: List[Track], track_path: str):
    with open(track_path, mode="w", encoding="utf-8") as _file:
        json.dump([track.to_dict() for track in tracks], _file)


if __name__ == "__main__":
    main()



"""Testing for the playlist_builder module.

Note: Core Spotify functionality is tested in the spotify-tools library.
These tests focus on DJ-Tools specific playlist building logic.
"""

from pathlib import Path
from unittest import mock

import pytest

from djtools.spotify.config import SubredditConfig
from djtools.spotify.playlist_builder import (
    async_spotify_playlists,
    spotify_playlist_from_upload,
    spotify_playlists,
)

from ..test_utils import MockOpen


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "playlist_subreddits",
    [[], [SubredditConfig(name="jungle")]],
)
@pytest.mark.parametrize("got_playlist_ids", [True, False])
@pytest.mark.parametrize("got_tracks", [True, False])
@mock.patch.object(Path, "exists", mock.Mock(return_value=True))
@mock.patch(
    "djtools.spotify.playlist_builder.get_subreddit_posts",
    new_callable=mock.AsyncMock,
    return_value=[
        [("track-id", "track name")],
        SubredditConfig(name="jungle"),
    ],
)
@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
@mock.patch("djtools.spotify.playlist_builder.get_reddit_client")
async def test_async_spotify_playlists(
    mock_reddit_client,
    _mock_spotify_client,
    mock_get_subreddit_posts,
    got_tracks,
    got_playlist_ids,
    playlist_subreddits,
    config,
):
    """Test for the async_spotify_playlists function."""
    if not got_tracks:
        mock_get_subreddit_posts.return_value[0] = []

    config.spotify.spotify_client_id = "test_client_id"
    config.spotify.spotify_client_secret = "test_client_secret"
    config.spotify.spotify_redirect_uri = "test_redirect_uri"
    config.spotify.spotify_playlist_subreddits = playlist_subreddits

    mock_reddit_client.return_value.close = mock.AsyncMock()

    with (
        mock.patch(
            "builtins.open",
            MockOpen(
                files=["spotify_playlists.yaml", ".praw.cache"],
                content='{"jungle": "some-id"}' if got_playlist_ids else "{}",
            ).open,
        ),
        mock.patch(
            "djtools.spotify.playlist_builder.populate_playlist",
            return_value={"jungle": "some-id"},
        ),
    ):
        await async_spotify_playlists(config)


@mock.patch("djtools.spotify.playlist_builder.filter_results")
@mock.patch("djtools.spotify.playlist_builder.populate_playlist")
@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
@mock.patch(
    "pyperclip.paste",
    return_value="""aweeeezy/Bass/2022-09-03: 5
 Brazil - A.M.C.mp3
 Endless Haze - Koherent.mp3
 Two Rangers - Two Rangers.mp3
 Under Pressure - Alix Perez, T-Man.mp3
 zoom.1 - Relativity Lounge, wicker's portal.mp3
aweeeezy/House/2022-09-03: 2
 Shirt - Cour T..mp3""",
)
def test_spotify_playlist_from_upload(
    _mock_paste,
    _mock_spotify_client,
    mock_populate,
    mock_filter,
    config,
    caplog,
):
    """Test for the spotify_playlist_from_upload function."""
    caplog.set_level("INFO")

    mock_filter.return_value = (
        {
            "id": "some_id",
            "name": "some_name",
            "artists": [{"name": "some_artist"}],
        },
        100,
    )
    mock_populate.return_value = {"some-playlist": "some-id"}

    config.spotify.spotify_client_id = "test_client_id"
    config.spotify.spotify_client_secret = "test_client_secret"
    config.spotify.spotify_redirect_uri = "test_redirect_uri"
    config.spotify.spotify_playlist_from_upload = True

    with mock.patch(
        "builtins.open",
        MockOpen(files=["spotify_playlists.yaml"], content="{}").open,
    ):
        spotify_playlist_from_upload(config)

    for rec in caplog.records:
        assert rec.message.startswith("Matched")


@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
@mock.patch(
    "pyperclip.paste",
    return_value="""aweeeezy/House/2022-09-03: 2
 UNKNOWN - 1 - Unknown Artist.mp3""",
)
def test_spotify_playlist_from_upload_handles_file_with_multiple_dashes(
    _mock_paste,
    _mock_spotify_client,
    config,
    caplog,
):
    """Test for the spotify_playlist_from_upload function."""
    caplog.set_level("WARNING")

    config.spotify.spotify_client_id = "test_client_id"
    config.spotify.spotify_client_secret = "test_client_secret"
    config.spotify.spotify_redirect_uri = "test_redirect_uri"
    config.spotify.spotify_playlist_from_upload = True

    with mock.patch(
        "builtins.open",
        MockOpen(files=["spotify_playlists.yaml"], content="{}").open,
    ):
        spotify_playlist_from_upload(config)

    assert (
        caplog.records[0].message
        == "UNKNOWN - 1 - Unknown Artist.mp3 is not a valid file"
    )


@mock.patch("djtools.spotify.playlist_builder.filter_results")
@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
def test_spotify_playlist_from_upload_handles_non_match(
    _mock_spotify_client,
    mock_filter,
    config,
    caplog,
):
    """Test for the spotify_playlist_from_upload function."""
    caplog.set_level("WARNING")
    title = "Under Pressure"
    artist = "Alix Perez, T-Man"

    mock_filter.return_value = ({}, 100)
    config.spotify.spotify_playlist_from_upload = True

    with (
        mock.patch(
            "builtins.open",
            MockOpen(files=["spotify_playlists.yaml"], content="{}").open,
        ),
        mock.patch(
            "pyperclip.paste",
            return_value=f"""aweeeezy/Bass/2022-09-03: 5
 {title} - {artist}.mp3""",
        ),
    ):
        spotify_playlist_from_upload(config)

    assert caplog.records[0].message == (
        f"Could not find a match for {title} - {artist}"
    )


@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
def test_spotify_playlist_from_upload_handles_spotify_exception(
    mock_spotify_client,
    config,
    caplog,
):
    """Test for the spotify_playlist_from_upload function."""
    caplog.set_level("ERROR")
    title = "Under Pressure"
    artist = "Alix Perez, T-Man"

    mock_spotify_client.return_value.search.side_effect = Exception(
        "API Error"
    )
    config.spotify.spotify_playlist_from_upload = True
    config.spotify.spotify_client_id = "test_client_id"
    config.spotify.spotify_client_secret = "test_client_secret"
    config.spotify.spotify_redirect_uri = "test_redirect_uri"

    with (
        mock.patch(
            "builtins.open",
            MockOpen(files=["spotify_playlists.yaml"], content="{}").open,
        ),
        mock.patch(
            "pyperclip.paste",
            return_value=f"""aweeeezy/Bass/2022-09-03: 5
 {title} - {artist}.mp3""",
        ),
    ):
        spotify_playlist_from_upload(config)

    assert caplog.records[0].message.startswith(
        f'Error searching for "{title} - {artist}"'
    )


@mock.patch("pyperclip.paste", return_value="")
def test_spotify_playlist_from_upload_raises_runtimeerror(_mock_paste, config):
    """Test for the spotify_playlist_from_upload function."""
    config.spotify.spotify_playlist_from_upload = True
    with pytest.raises(
        RuntimeError,
        match="Generating a Spotify playlist from an upload requires output "
        "from an upload_music Discord webhook to be copied to the "
        "system's clipboard",
    ):
        spotify_playlist_from_upload(config)


def test_spotify_playlists(config):
    """Test for the spotify_playlists function."""
    with mock.patch(
        "djtools.spotify.playlist_builder.async_spotify_playlists"
    ) as mock_async_spotify_playlists:
        mock_async_spotify_playlists.return_value = lambda x: None
        spotify_playlists(config)
        assert mock_async_spotify_playlists.called


@mock.patch("djtools.spotify.playlist_builder.filter_results")
@mock.patch("djtools.spotify.playlist_builder.populate_playlist")
@mock.patch("djtools.spotify.playlist_builder.get_spotify_client")
@mock.patch(
    "pyperclip.paste",
    return_value="""aweeeezy/Bass/2022-09-03: 1
 A.M.C - Brazil.mp3""",
)
def test_spotify_playlist_from_upload_artist_first(
    _mock_paste,
    _mock_spotify_client,
    mock_populate,
    mock_filter,
    config,
    caplog,
):
    """Test for spotify_playlist_from_upload with artist_first=True."""
    caplog.set_level("INFO")

    mock_filter.return_value = (
        {
            "id": "some_id",
            "name": "Brazil",
            "artists": [{"name": "A.M.C"}],
        },
        100,
    )
    mock_populate.return_value = {"some-playlist": "some-id"}

    config.spotify.spotify_client_id = "test_client_id"
    config.spotify.spotify_client_secret = "test_client_secret"
    config.spotify.spotify_redirect_uri = "test_redirect_uri"
    config.spotify.spotify_playlist_from_upload = True
    config.sync.artist_first = True

    with mock.patch(
        "builtins.open",
        MockOpen(files=["spotify_playlists.yaml"], content="{}").open,
    ):
        spotify_playlist_from_upload(config)

    # Verify filter_results was called with track and artist swapped
    # (artist_first means the file is "Artist - Track" so we swap to "Track - Artist")
    mock_filter.assert_called()
    call_args = mock_filter.call_args
    # The third and fourth args should be track and artist (after swapping)
    assert call_args[0][3] == "Brazil"  # track (was artist position)
    assert call_args[0][4] == "A.M.C"  # artist (was track position)



"""This is a script for sorting the non-genre tags of a collection in
alphabetical order.
"""

# pylint: disable=protected-access
from argparse import ArgumentParser
import re

from djtools.configs import build_config
from djtools.collection.platform_registry import PLATFORM_REGISTRY


TAG_ORDERINGS = (
    (
        "Aggro",
        "Atmospheric",
        "Bounce",
        "Dark",
        "Deep",
        "Gangsta",
        "Groovy",
        "Heavy",
        "Hypnotic",
        "Melancholy",
        "Melodic",
        "Rave",
        "Strange",
        "Uplifting",
    ),
    ("Flute", "Guitar", "Horn", "Piano", "Strings", "Vocal"),
    ("Loop", "Scratch"),
    ("DELETE",),
)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--config", required=True, help="Path to a config.yaml"
    )
    parser.add_argument(
        "--collection", required=False, help="Path to a collection"
    )
    args = parser.parse_args()

    # Load config and collection and get a dict of tracks keyed by location.
    config = build_config(args.config)
    collection = PLATFORM_REGISTRY[config.platform]["collection"](
        path=args.collection or config.collection_path
    )
    playlist_class = PLATFORM_REGISTRY[config.platform]["playlist"]

    tracks = collection.get_tracks()
    tag_sets = [set(tags) for tags in TAG_ORDERINGS]
    MY_TAGS_REGEX = r"(?<=\/\*).*(?=\*\/)"
    affected_tracks = {}

    for track_id, track in tracks.items():
        all_tags = track.get_tags()
        genre_tags = set(track.get_genre_tags())
        track_tag_set = set(all_tags).difference(genre_tags)
        new_tags = []
        for tag_order, tag_set in zip(TAG_ORDERINGS, tag_sets):
            new_tags.extend(sorted(track_tag_set.intersection(tag_set)))

        tags_before = [tag for tag in all_tags if tag not in genre_tags]
        if tags_before != new_tags:
            print(
                f"{track._Name} - {track._Artist}\n{tags_before}\n{new_tags}\n"
            )
            track._Comments = re.sub(
                MY_TAGS_REGEX,
                " " + " / ".join(new_tags) + " ",
                track._Comments,
            )
            affected_tracks[track_id] = track

    collection.add_playlist(
        playlist_class.new_playlist(
            name="Tracks with sorted comments", tracks=affected_tracks
        )
    )
    collection.serialize()

    print(f"{len(affected_tracks)} out of {len(tracks)} tracks were affected.")



"""This is a script for renaming files in a collection so they exist at the
desired location following this convention:
        '<usb>/DJ Music/<user>/<path-insert>/<filename>'

Users provide <usb> and <user> via the CLI. <path_insert> may also be provided
(if you want all your tracks to be in the same directory), otherwise it is
inferred from the common path between tracks of the collection. <filename> will
be the same filename for tracks unless '--infer-file-names' is provided which
enables inferring the filename using the tracks' 'title' and 'artist' ID3 tags.

The 'analyze_tracks' function is run first which will gather some statistics
regarding how many tracks don't adhere to the desired structure, how many have
a sample rate below 44.1 kHz, and how many couldn't be opened.

Running this script without setting '--actually-move-stuff' will only create an
output_rekordbox.xml which you may use to preview the modifications made to
your collection. Once you're satisfied with the analysis of your collection and
the output_rekordbox.xml, you can run with this flag set to actually relocate
your files.

NOTE: this is a move and not a copy!

Also note that you might want to preemptively delete all the tracks in your
collection (after backing everything up, of course) so that you don't have to
worry about deduplicating tracks in your collection once you import them from
the new locations.

Example usage:

Say I have all my music located under "/Users/aweeeezy/Desktop/music/" and there's
a pre-existing directory structure there that I want to preserve. I also want
to use the filenames as they are already:

    python scripts/repair/rename_files.py \
        --collection /path/to/my/rekordbox.xml \
        --usb "/Volumes/AWEEEEZY/" \
        --user aweeeezy \
        --actually-move-stuff

Instead, say we have the same scenario above except all the files in my
collection are techno so I want to collapse them into a flat directory called
"Techno":

    python scripts/repair/rename_files.py \
        --collection /path/to/my/rekordbox.xml \
        --usb "/Volumes/AWEEEEZY/" \
        --user aweeeezy \
        --actually-move-stuff \
        --path-insert "Techno"

Keep in mind that you can run this script multiple times using tailored XML
files and different arguments to '--path-insert' in order to move your library
in batches. For example, you can use the 'djtools.collection' API to filter
tracks in your collection by genre tags, substrings in their paths, or whatever
really, creating new XML files that you can run with this script to do targeted
relocations.

See the reference section for the Collection package in the docs
or ask me for more details:
https://a-rich.github.io/DJ-Tools/reference/collection/

Also checkout the developer docs on the collection API:
https://a-rich.github.io/DJ-Tools-dev-docs/tutorials/developer_docs/collections_api/

Finally, let's say we want to move everything directly under my user folder and
also rename all the files using the 'title' and 'artist' ID3 tags -- providing
an empty string for '--path-insert' will disable inferring the path suffix from
the common path of tracks in your collection:

    python scripts/repair/rename_files.py \
        --collection /path/to/my/rekordbox.xml \
        --usb "/Volumes/AWEEEEZY/" \
        --user aweeeezy \
        --actually-move-stuff \
        --path-insert "" \
        --infer-file-names
"""

# pylint: disable=import-error,redefined-outer-name,no-member
from argparse import ArgumentParser
from concurrent.futures import as_completed, ThreadPoolExecutor
import json
import os
from pathlib import Path
import shutil
from typing import Dict, Optional

import eyed3
from pydub import AudioSegment
from tqdm import tqdm

from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.rekordbox_track import RekordboxTrack
from djtools.configs.cli_args import convert_to_paths


eyed3.log.setLevel("ERROR")


def analyze_tracks(collection: RekordboxCollection, usb: Path) -> str:
    """Determines statistics of a collection.

    Tracks are noted if:
        - their paths don't have a prefix matching the desired path
        - they don't exist at the path specified
        - they have a sample rate lower than 44.1kHz

    Args:
        collection: Collection object.
        usb: Path to a location where all tracks will be relocated.

    Returns:
        Common path across tracks in the collection.
    """

    def thread(track: RekordboxTrack, data: Dict, usb: Path) -> Path:
        """Threaded process for processing tracks.

        Args:
            track: Track object.
            data: Dictionary to store track statistics.
            usb: Path to a location where all tracks will be relocated.

        Returns:
            Path of the track.
        """
        loc = track.get_location()
        if not str(loc).startswith(str(usb)):
            data["tracks_needing_relocation"].append(str(loc))
        try:
            audio = AudioSegment.from_file(loc)
        except FileNotFoundError:
            data["tracks_unable_to_be_opened"].append(str(loc))
            return loc
        if audio.frame_rate < 44_100:
            data["tracks_with_low_sample_rate"].append(str(loc))

        return loc

    # Thread pool to process tracks in the collection.
    tracks = list(collection.get_tracks().values())
    data = {
        "tracks_needing_relocation": [],
        "tracks_with_low_sample_rate": [],
        "tracks_unable_to_be_opened": [],
    }
    locations = []
    with tqdm(total=len(tracks)) as pbar:
        with ThreadPoolExecutor(max_workers=32) as pool:
            futures = [
                pool.submit(thread, track, data, usb) for track in tracks
            ]
            for future in as_completed(futures):
                locations.append(future.result())
                pbar.update()
            locations = list(filter(None, locations))

    if not locations:
        print(
            "WARNING: none of your tracks were able to be opened...make sure "
            "whatever file system you store your music on is mounted."
        )
        data["common_path"] = ""
    else:
        data["common_path"] = os.path.commonpath(locations) + os.path.sep

    # Print statistics about tracks.
    for key, value in data.items():
        print(f"{key}: {len(value) if isinstance(value, list) else value}")

    # Write statistics to a JSON file.
    with open(
        "track_analysis_data.json",
        mode="w",
        encoding="utf-8",
    ) as _file:
        json.dump(data, _file)

    return data["common_path"]


def move_files(
    collection: RekordboxCollection,
    usb: Path,
    user: str,
    common_path: str,
    path_insert: Optional[str] = None,
    infer_file_names: bool = False,
    actually_move_stuff: bool = False,
):
    """Moves tracks in a collection to a new location.

    Given a Collection object, this function will move all the tracks to a path
    following the convention:
        '<usb>/DJ Music/<user>/<path_insert>/<filename>'

    Users provide <usb> and <user> via the CLI. <path_insert> may also be
    provided (if you want all your tracks to be in the same directory),
    otherwise it is inferred from the common path between tracks of the
    collection. <filename> will be the same filename for tracks unless
    'infer_file_names' is provided which will infer the filename using the
    tracks' 'title' and 'artist' ID3 tags.

    Args:
        collection: Collection object.
        usb: Path to a location where all tracks will be relocated.
        user: User's name to use in new track paths.
        common_path: Path common to all tracks in the collection. When
            determining new track locations, old locations will be split on
            this to infer a path suffix. Unless '--path-insert' is used, the
            inferred suffix will be inserted into the new path locations
            effectively preserving your directory structure.
        path_insert: Path to insert after "user"
        infer_file_names: If this is set, the filenames as the new track paths
            will be inferred from the 'title' and 'artist' ID3 tags.
        actually_move_stuff: If set, this will move files around.
    """

    def thread(
        track: RekordboxTrack,
        usb: Path,
        user: str,
        common_path: str,
        infer_file_names: bool,
        actually_move_stuff: bool,
        data: Dict,
    ):
        """Threaded process for processing tracks.

        Args:
            track: Track object from a Collection.
            usb: Path to a location where all tracks will be relocated.
            user: User's name to use in new track paths.
            common_path: Path common to all tracks in the collection. When
                determining new track locations, old locations will be split on
                this to infer a path suffix. Unless '--path-insert' is used,
                the inferred suffix will be inserted into the new path
                locations effectively preserving your directory structure.
            actually_move_stuff: If set, this will move files around.
            data: Dictionary to store track statistics.
        """
        loc = track.get_location()

        # Skip files that don't exist.
        if not loc.exists():
            data["tracks_unable_to_be_opened"].append(str(loc))
            if actually_move_stuff:
                return

        # Ignore track already under '<usb>/DJ Music/' as they are assumed to
        # be correct.
        if str(loc).startswith(str(usb / "DJ Music")):
            data["tracks_already_at_destination"].append(str(loc))
            return

        # Determine the part of the path between '<usb>/DJ Music/<user>/' and
        # the filename.
        if path_insert is not None:
            path_suffix = path_insert
        else:
            path_suffix = str(loc.parent).split(common_path, maxsplit=1)[-1]

        # Either use the existing filename or infer a new one using the ID3
        # tags for 'title' and 'artist'.
        if infer_file_names:
            track_id3 = eyed3.load(loc)
            title = getattr(track_id3.tag, "title")
            artist = getattr(track_id3.tag, "artist")
            if title and artist:
                filename = f"{title} - {artist}{loc.suffix}"
            else:
                data["tracks_missing_tags"].append(str(loc))
                filename = loc.name
        else:
            filename = loc.name

        # Move the track and update the location of it in the collection.
        new_loc = usb / "DJ Music" / user / path_suffix / filename
        if actually_move_stuff:
            if not new_loc.parent.exists():
                new_loc.parent.mkdir(parents=True, exist_ok=True)
            shutil.move(loc, new_loc)
        track.set_location(new_loc)

    # Thread pool to process tracks in the collection.
    data = {
        "tracks_missing_tags": [],
        "tracks_already_at_destination": [],
        "tracks_unable_to_be_opened": [],
    }
    tracks = list(collection.get_tracks().values())
    with tqdm(total=len(tracks)) as pbar:
        with ThreadPoolExecutor(max_workers=32) as pool:
            futures = [
                pool.submit(
                    thread,
                    track,
                    usb,
                    user,
                    common_path,
                    infer_file_names,
                    actually_move_stuff,
                    data,
                )
                for track in tracks
            ]
            for future in as_completed(futures):
                future.result()
                pbar.update()

    # Serialize the updated collection to a new XML file.
    collection.serialize(path="output_rekordbox.xml")

    # Print statistics about tracks.
    for key, value in data.items():
        print(f"{key}: {len(value)}")

    # Write statistics to a JSON file.
    with open(
        "track_move_data.json",
        mode="w",
        encoding="utf-8",
    ) as _file:
        json.dump(data, _file)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--collection",
        required=True,
        type=convert_to_paths,
        help="Path to a Rekordbox XML file.",
    )
    parser.add_argument(
        "--usb",
        required=True,
        type=convert_to_paths,
        help=(
            'Path to a "DJ Music" directory where you\'d like all your music '
            "to be."
        ),
    )
    parser.add_argument(
        "--user",
        required=True,
        type=str,
        help=(
            "Your beatcloud username. This MUST be consistent with what may "
            "already exist in the beatcloud."
        ),
    )
    parser.add_argument(
        "--path-insert",
        type=str,
        help=(
            "Optional path to insert between '--user' and the basename of "
            "you files. If not provided, the part of the path between the "
            "basename and the prefix common to all tracks will be used."
        ),
    )
    parser.add_argument(
        "--infer-file-names",
        action="store_true",
        help=(
            "If set, this will change the filename to be "
            "{title} - {artist}.{ext} where 'title' and 'artist' are taken "
            "from the track's ID3 tags."
        ),
    )
    parser.add_argument(
        "--actually-move-stuff",
        action="store_true",
        help="If set, this will move files around.",
    )
    args = parser.parse_args()

    # Deserialize a Rekordbox XML file.
    collection = RekordboxCollection(args.collection)

    # This just extracts location info to see (a) what the common path is for
    # tracks in your collection and (b) what the list of tracks not located
    # under 'args.usb' is.
    #
    # Nothing destructive occurs.
    common_path = analyze_tracks(collection, args.usb)

    # This function, however, is destructive.
    #
    # It will move every file not under 'args.usb' to
    # '<args.usb>/DJ Music/<args.user>/' and modify the location in your
    # collection as well, enabling you to re-import these tracks.
    move_files(
        collection,
        args.usb,
        args.user,
        common_path,
        path_insert=args.path_insert,
        infer_file_names=args.infer_file_names,
        actually_move_stuff=args.actually_move_stuff,
    )



"""Testing for the helpers module.

Note: Core Spotify functionality is tested in the spotify-tools library.
These tests focus on DJ-Tools specific integration and helper functions.
"""

import asyncio
from pathlib import Path
from unittest import mock

import pytest
import yaml

from djtools.spotify.helpers import (
    _build_new_playlist,
    _catch,
    _filter_tracks,
    _fuzzy_match,
    _parse_title,
    _process,
    _track_name_too_similar,
    _update_existing_playlist,
    filter_results,
    get_playlist_ids,
    get_reddit_client,
    get_spotify_client,
    get_subreddit_posts,
    populate_playlist,
    write_playlist_ids,
)

from ..test_utils import mock_exists, MockOpen


async def _aiter(obj, num_subs):
    """Helper function for mocking asyncpraw."""
    for _ in range(num_subs):
        yield obj
        await asyncio.sleep(0.1)


@mock.patch("djtools.spotify.helpers.Client")
def test_build_new_playlist(mock_client):
    """Test for the _build_new_playlist function."""
    mock_client.user_playlist_create.return_value = {"id": "test_id"}
    new_tracks = [("test_id", "track title - artist name")]
    ret = _build_new_playlist(mock_client, "test_user", "r/techno", new_tracks)
    assert isinstance(ret, dict)
    assert ret == {"id": "test_id"}
    mock_client.user_playlist_create.assert_called_once()
    mock_client.playlist_add_items.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize("message", ["", "oops"])
async def test_catch(message, caplog):
    """Test for the _catch function."""
    exc = ZeroDivisionError("You can't divide by zero!")

    class Generator:
        """Dummy async generator class."""

        def __init__(self):
            self._iters = 2
            self._i = 0

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self._i >= self._iters:
                raise StopAsyncIteration
            self._i += 1
            await asyncio.sleep(0.1)
            if self._i % 2 == 0:
                raise exc
            return self._i

    caplog.set_level("WARNING")
    _ = [x async for x in _catch(Generator(), message=message)]
    assert caplog.records[0].message == (
        f"{message}: {str(exc)}" if message else str(exc)
    )


@mock.patch(
    "builtins.open",
    MockOpen(
        files=["spotify_playlists.yaml"],
        content="playlist: playlist-id",
    ).open,
)
@pytest.mark.parametrize(
    "config_exists, expected",
    [(True, {"playlist": "playlist-id"}), (False, {})],
)
def test_get_playlist_ids(config_exists, expected):
    """Test for the get_playlist_ids function."""
    with mock.patch(
        "djtools.collection.config.Path.exists",
        lambda path: mock_exists(
            [("spotify_playlists.yaml", config_exists)],
            path,
        ),
    ):
        playlist_ids = get_playlist_ids()
    assert isinstance(playlist_ids, dict)
    assert playlist_ids == expected


@mock.patch("djtools.spotify.helpers.praw.Reddit")
def test_get_reddit_client(mock_reddit, config):
    """Test for the get_reddit_client function."""
    config.spotify.reddit_client_id = "test_client_id"
    config.spotify.reddit_client_secret = "test_client_secret"
    config.spotify.reddit_user_agent = "test_user_agent"
    get_reddit_client(config)
    mock_reddit.assert_called_once()


@pytest.mark.parametrize("is_spotify_config", [True, False])
@mock.patch("djtools.spotify.helpers.Client")
def test_get_spotify_client(mock_client, is_spotify_config, config):
    """Test for the get_spotify_client function."""
    if is_spotify_config:
        config = config.spotify

    config.spotify_client_id = "test_client_id"
    config.spotify_client_secret = "test_client_secret"
    config.spotify_redirect_uri = "test_redirect_uri"

    get_spotify_client(config)
    mock_client.assert_called_once()


@pytest.mark.parametrize(
    "title",
    [
        "Arctic Oscillations by Fanu",
        "Arctic Oscillations - Fanu",
        "Arctic Oscillations - Fanu (this track is cool)",
        "Arctic Oscillations - Fanu [love this track]",
        "Fanu - Arctic Oscillations",
        "A submission title that doesn't include the artist or track info",
    ],
)
def test_parse_title(title):
    """Test for the _parse_title function."""
    split_chars = ["[", "("]
    ret = _parse_title(title)
    assert isinstance(ret, list)
    if " - " in title or " by " in title:
        assert all(x for x in ret)
        if any(x in title for x in split_chars):
            assert all(x not in r for x in split_chars for r in ret)
    else:
        assert not any(x for x in ret)


@pytest.mark.parametrize("playlist_ids", [{}, {"playlist": "id"}])
@pytest.mark.parametrize("tracks", [[], [("id", "title - artist")]])
@mock.patch("djtools.spotify.helpers.Client")
def test_populate_playlist(
    mock_client,
    playlist_ids,
    tracks,
    caplog,
):
    """Test for the populate_playlist function."""
    ret_val = {
        "name": "playlist",
        "external_urls": {"spotify": "https://test-url.com"},
        "id": "test-id",
    }
    mock_client.playlist.return_value = ret_val
    mock_client.user_playlist_create.return_value = ret_val
    caplog.set_level("INFO")

    with mock.patch(
        "djtools.spotify.helpers._update_existing_playlist",
        return_value=ret_val,
    ) as mock_update:
        populate_playlist(
            playlist_name="playlist",
            playlist_ids=dict(playlist_ids),
            spotify_username="test",
            spotify=mock_client,
            tracks=tracks,
            playlist_limit=50,
        )

    if playlist_ids and tracks:
        assert mock_update.call_count == 1
    elif tracks:
        assert "Unable to get ID for playlist" in caplog.text
    elif playlist_ids:
        mock_client.playlist.assert_called_once()

    if not (playlist_ids or tracks):
        assert mock_update.call_count == 0


@pytest.mark.parametrize(
    "url",
    [
        "https://open.spotify.com/track/1lps8esDJ9M6rG3HBjhuux",
        "https://some-other-url.com/some_id",
    ],
)
@mock.patch("djtools.spotify.helpers.Client")
@mock.patch("djtools.spotify.helpers.praw.models.Submission")
def test_process(mock_submission, mock_client, url):
    """Test for the _process function."""
    title = "Arctic Oscillations - Fanu"
    mock_submission.url = url
    mock_submission.title = title

    with mock.patch(
        "djtools.spotify.helpers._fuzzy_match", return_value=(url, title)
    ):
        ret = _process(mock_submission, mock_client, 50)

    assert isinstance(ret, tuple)
    assert ret == (url, title)


@pytest.mark.parametrize(
    "playlist_track_names",
    [
        {"Arctic Oscillations - Fanu"},
        {"Not a Match - Some Artist"},
    ],
)
def test_track_name_too_similar(playlist_track_names, caplog):
    """Test for the _track_name_too_similar function."""
    caplog.set_level("WARNING")
    track = "Arctic Oscillations - Fanu"
    ret = _track_name_too_similar(track, playlist_track_names)
    if any("Not a Match" in x for x in playlist_track_names):
        assert not ret
    else:
        assert ret
        assert "too similar" in caplog.text


def test_write_playlist_ids():
    """Test for the write_playlist_ids function."""
    ids_path = (
        Path(__file__).parent.parent.parent
        / "src"
        / "djtools"
        / "configs"
        / "spotify_playlists.yaml"
    )

    # Write a playlist to the config
    test_data = {"playlist": "playlist-id", "another": "another-id"}
    write_playlist_ids(test_data)

    with open(ids_path, mode="r", encoding="utf-8") as _file:
        data = yaml.load(_file, Loader=yaml.FullLoader)

    assert data == test_data

    # Clean up
    ids_path.unlink()


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_with_matches(mock_client):
    """Test filter_results returns best match."""
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "Test Song",
                    "artists": [{"name": "Test Artist"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": None,
        }
    }

    track, score = filter_results(
        mock_client, results, 70.0, "Test Song", "Test Artist"
    )

    assert track.get("id") == "track1"
    assert score > 0


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_no_matches(mock_client):
    """Test filter_results returns empty when no matches."""
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "Completely Different",
                    "artists": [{"name": "Unknown"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": None,
        }
    }

    track, score = filter_results(
        mock_client, results, 90.0, "Test Song", "Test Artist"
    )

    assert track == {}
    assert score == 0.0


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_with_pagination(mock_client):
    """Test filter_results handles pagination."""
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "First Track",
                    "artists": [{"name": "Artist"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": "https://api.spotify.com/next",
        }
    }
    mock_client.next.return_value = {
        "tracks": {
            "items": [
                {
                    "id": "track2",
                    "name": "Test Song",
                    "artists": [{"name": "Test Artist"}],
                    "uri": "spotify:track:track2",
                }
            ],
            "next": None,
        }
    }

    track, score = filter_results(
        mock_client, results, 70.0, "Test Song", "Test Artist"
    )

    assert track.get("id") == "track2"
    assert score > 0
    mock_client.next.assert_called()


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_pagination_exception(mock_client, caplog):
    """Test filter_results handles pagination errors."""
    caplog.set_level("WARNING")
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "Test Song",
                    "artists": [{"name": "Test Artist"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": "https://api.spotify.com/next",
        }
    }
    mock_client.next.side_effect = Exception("API Error")

    track, _score = filter_results(
        mock_client, results, 70.0, "Test Song", "Test Artist"
    )

    assert track.get("id") == "track1"
    assert "Failed to get next tracks" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_pagination_empty_response(mock_client):
    """Test filter_results handles empty pagination response."""
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "Test Song",
                    "artists": [{"name": "Test Artist"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": "https://api.spotify.com/next",
        }
    }
    mock_client.next.return_value = None

    track, _ = filter_results(
        mock_client, results, 70.0, "Test Song", "Test Artist"
    )

    assert track.get("id") == "track1"


@mock.patch("djtools.spotify.helpers.Client")
def test_filter_results_pagination_items_key(mock_client):
    """Test filter_results handles 'items' key in pagination response."""
    results = {
        "tracks": {
            "items": [
                {
                    "id": "track1",
                    "name": "First",
                    "artists": [{"name": "Artist"}],
                    "uri": "spotify:track:track1",
                }
            ],
            "next": "https://api.spotify.com/next",
        }
    }
    # Simulating response where items is at top level instead of tracks.items
    mock_client.next.return_value = {
        "items": [
            {
                "id": "track2",
                "name": "Test Song",
                "artists": [{"name": "Test Artist"}],
                "uri": "spotify:track:track2",
            }
        ],
        "next": None,
    }

    _, score = filter_results(
        mock_client, results, 70.0, "Test Song", "Test Artist"
    )

    assert score > 0


def test_filter_tracks():
    """Test _filter_tracks function."""
    tracks = [
        {
            "id": "track1",
            "name": "Test Song",
            "artists": [{"name": "Test Artist"}],
            "uri": "spotify:track:track1",
        },
        {
            "id": "track2",
            "name": "Another Track",
            "artists": [{"name": "Other Artist"}],
            "uri": "spotify:track:track2",
        },
    ]

    results = _filter_tracks(tracks, 70.0, "Test Song", "Test Artist")
    assert len(results) == 1
    assert results[0][0]["id"] == "track1"
    assert results[0][1] > 0


def test_filter_tracks_no_matches():
    """Test _filter_tracks returns empty when no matches."""
    tracks = [
        {
            "id": "track1",
            "name": "Completely Different",
            "artists": [{"name": "Unknown"}],
            "uri": "spotify:track:track1",
        }
    ]

    results = _filter_tracks(tracks, 90.0, "Test Song", "Test Artist")
    assert results == []


@mock.patch("djtools.spotify.helpers.search_track_fuzzy")
@mock.patch("djtools.spotify.helpers.Client")
def test_fuzzy_match_success(mock_client, mock_search):
    """Test _fuzzy_match returns track when found."""
    mock_track = mock.MagicMock()
    mock_track.id = "track123"
    mock_track.name = "Test Song"
    mock_artist = mock.MagicMock()
    mock_artist.name = "Test Artist"
    mock_track.artists = [mock_artist]

    mock_result = mock.MagicMock()
    mock_result.track = mock_track
    mock_search.return_value = mock_result

    result = _fuzzy_match(mock_client, "Test Song - Test Artist", 70.0)

    assert result is not None
    assert result[0] == "track123"
    assert "Test Song" in result[1]


@mock.patch("djtools.spotify.helpers.search_track_fuzzy")
@mock.patch("djtools.spotify.helpers.Client")
def test_fuzzy_match_no_match(mock_client, mock_search):
    """Test _fuzzy_match returns None when no match found."""
    mock_search.return_value = None

    result = _fuzzy_match(mock_client, "Test Song - Test Artist", 70.0)

    assert result is None


@mock.patch("djtools.spotify.helpers.Client")
def test_fuzzy_match_unparseable_title(mock_client):
    """Test _fuzzy_match returns None for unparseable titles."""
    result = _fuzzy_match(mock_client, "No delimiter here", 70.0)
    assert result is None


@mock.patch("djtools.spotify.helpers.search_track_fuzzy")
@mock.patch("djtools.spotify.helpers.Client")
def test_fuzzy_match_exception(mock_client, mock_search, caplog):
    """Test _fuzzy_match handles exceptions gracefully."""
    caplog.set_level("ERROR")
    mock_search.side_effect = Exception("API Error")

    result = _fuzzy_match(mock_client, "Test Song - Test Artist", 70.0)

    assert result is None
    assert "Error searching" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_add_tracks(mock_client, caplog):
    """Test _update_existing_playlist adds new tracks."""
    caplog.set_level("INFO")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "existing1",
                        "name": "Existing Track",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:existing1",
                    }
                }
            ],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data

    new_tracks = [("new_track_id", "New Track - New Artist")]

    result = _update_existing_playlist(
        mock_client, "playlist123", new_tracks, 50, 0
    )

    assert result["id"] == "playlist123"
    mock_client.playlist_add_items.assert_called_once()
    assert "1 new tracks added" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_duplicate_id(mock_client, caplog):
    """Test _update_existing_playlist skips duplicate track IDs."""
    caplog.set_level("WARNING")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "existing1",
                        "name": "Existing Track",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:existing1",
                    }
                }
            ],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data

    new_tracks = [("existing1", "Existing Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    mock_client.playlist_add_items.assert_not_called()
    assert "already in the playlist" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_resolves_url(mock_client, caplog):
    """Test _update_existing_playlist resolves Spotify URLs to track IDs."""
    caplog.set_level("INFO")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data
    mock_client.track.return_value = {
        "id": "resolved_id",
        "name": "Resolved Track",
        "artists": [{"name": "Artist"}],
    }

    new_tracks = [
        (
            "https://open.spotify.com/track/abc123",
            "Some Track - Some Artist",
        )
    ]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    mock_client.track.assert_called_once()
    mock_client.playlist_add_items.assert_called_once()


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_removes_old_tracks(mock_client, caplog):
    """Test _update_existing_playlist removes old tracks when limit exceeded."""
    caplog.set_level("INFO")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "old1",
                        "name": "Old Track",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:old1",
                    }
                }
            ],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data

    new_tracks = [("new1", "New Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 1, 0)

    mock_client.playlist_remove_specific_occurrences_of_items.assert_called()
    mock_client.playlist_add_items.assert_called_once()
    assert "1 old tracks removed" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_verbose_logging(mock_client, caplog):
    """Test _update_existing_playlist logs track details when verbose."""
    caplog.set_level("INFO")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data

    new_tracks = [("new1", "New Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 1)

    assert "New Track - Artist" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_no_changes(mock_client, caplog):
    """Test _update_existing_playlist logs when no changes made."""
    caplog.set_level("INFO")
    playlist_data = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "existing1",
                        "name": "Existing Track",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:existing1",
                    }
                }
            ],
            "next": None,
        },
    }
    mock_client.playlist.return_value = playlist_data

    new_tracks = [("existing1", "Existing Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    assert "No tracks added or removed" in caplog.text


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_pagination(mock_client):
    """Test _update_existing_playlist handles pagination."""
    page1 = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "track1",
                        "name": "Track 1",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:track1",
                    }
                }
            ],
            "next": "https://api.spotify.com/next",
        },
    }
    page2 = {
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "track2",
                        "name": "Track 2",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:track2",
                    }
                }
            ],
            "next": None,
        }
    }
    mock_client.playlist.return_value = page1
    mock_client.next.return_value = page2

    new_tracks = [("new_track", "New Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    mock_client.next.assert_called()
    mock_client.playlist_add_items.assert_called_once()


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_pagination_items_key(mock_client):
    """Test _update_existing_playlist handles pagination with items key."""
    page1 = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [
                {
                    "track": {
                        "id": "track1",
                        "name": "Track 1",
                        "artists": [{"name": "Artist"}],
                        "uri": "spotify:track:track1",
                    }
                }
            ],
            "next": "https://api.spotify.com/next",
        },
    }
    # Response with items at top level instead of tracks.items
    page2 = {
        "items": [
            {
                "track": {
                    "id": "track2",
                    "name": "Track 2",
                    "artists": [{"name": "Artist"}],
                    "uri": "spotify:track:track2",
                }
            }
        ],
        "next": None,
    }
    mock_client.playlist.return_value = page1
    mock_client.next.return_value = page2

    new_tracks = [("new_track", "New Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    mock_client.next.assert_called()


@mock.patch("djtools.spotify.helpers.Client")
def test_update_existing_playlist_pagination_error(mock_client, caplog):
    """Test _update_existing_playlist handles pagination errors."""
    caplog.set_level("ERROR")
    page1 = {
        "id": "playlist123",
        "name": "Test Playlist",
        "tracks": {
            "items": [],
            "next": "https://api.spotify.com/next",
        },
    }
    mock_client.playlist.return_value = page1
    mock_client.next.side_effect = Exception("API Error")

    new_tracks = [("new_track", "New Track - Artist")]

    _update_existing_playlist(mock_client, "playlist123", new_tracks, 50, 0)

    assert "Failed to get tracks" in caplog.text


class MockTqdm:
    """Mock tqdm that works both as iterator and context manager."""

    def __init__(
        self, iterable=None, **kwargs
    ):  # pylint: disable=unused-argument
        self.iterable = iterable

    def __iter__(self):
        return iter(self.iterable or [])

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def update(self, n=1):
        """Mock update method."""


@pytest.mark.asyncio
@mock.patch("tqdm.tqdm", MockTqdm)
@mock.patch("djtools.spotify.helpers.Client")
async def test_get_subreddit_posts(mock_client, config):
    """Test get_subreddit_posts function."""
    config.spotify.spotify_playlist_post_limit = 10
    config.spotify.spotify_playlist_fuzz_ratio = 70

    mock_reddit = mock.AsyncMock()
    mock_subreddit = mock.AsyncMock()
    mock_reddit.subreddit.return_value = mock_subreddit

    # Create mock submissions
    mock_submission = mock.MagicMock()
    mock_submission.id = "sub123"
    mock_submission.url = "https://open.spotify.com/track/abc123"
    mock_submission.title = "Test Track - Test Artist"

    mock_subreddit.hot = mock.MagicMock(
        return_value=_aiter(mock_submission, 1)
    )

    from djtools.spotify.config import SubredditConfig
    from djtools.spotify.enums import SubredditType

    subreddit_config = SubredditConfig(name="techno", type=SubredditType.HOT)

    tracks, sub = await get_subreddit_posts(
        mock_client, mock_reddit, subreddit_config, config, {}
    )

    assert sub == subreddit_config
    assert len(tracks) == 1


@pytest.mark.asyncio
@mock.patch("tqdm.tqdm", MockTqdm)
@mock.patch("djtools.spotify.helpers.Client")
async def test_get_subreddit_posts_with_cache(mock_client, config, caplog):
    """Test get_subreddit_posts skips cached submissions."""
    caplog.set_level("INFO")
    config.spotify.spotify_playlist_post_limit = 10
    config.spotify.spotify_playlist_fuzz_ratio = 70

    mock_reddit = mock.AsyncMock()
    mock_subreddit = mock.AsyncMock()
    mock_reddit.subreddit.return_value = mock_subreddit

    mock_submission = mock.MagicMock()
    mock_submission.id = "cached_sub"
    mock_submission.url = "https://open.spotify.com/track/abc123"
    mock_submission.title = "Test Track"

    mock_subreddit.hot = mock.MagicMock(
        return_value=_aiter(mock_submission, 1)
    )

    from djtools.spotify.config import SubredditConfig
    from djtools.spotify.enums import SubredditType

    subreddit_config = SubredditConfig(name="techno", type=SubredditType.HOT)
    praw_cache = {"cached_sub": True}

    tracks, _ = await get_subreddit_posts(
        mock_client, mock_reddit, subreddit_config, config, praw_cache
    )

    assert len(tracks) == 0
    assert "No new submissions" in caplog.text


@pytest.mark.asyncio
@mock.patch("tqdm.tqdm", MockTqdm)
@mock.patch("djtools.spotify.helpers.Client")
async def test_get_subreddit_posts_top_type(mock_client, config):
    """Test get_subreddit_posts with TOP subreddit type."""
    config.spotify.spotify_playlist_post_limit = 10
    config.spotify.spotify_playlist_fuzz_ratio = 70

    mock_reddit = mock.AsyncMock()
    mock_subreddit = mock.AsyncMock()
    mock_reddit.subreddit.return_value = mock_subreddit

    mock_submission = mock.MagicMock()
    mock_submission.id = "sub123"
    mock_submission.url = "https://open.spotify.com/track/abc123"
    mock_submission.title = "Test Track"

    mock_subreddit.top = mock.MagicMock(
        return_value=_aiter(mock_submission, 1)
    )

    from djtools.spotify.config import SubredditConfig
    from djtools.spotify.enums import SubredditPeriod, SubredditType

    subreddit_config = SubredditConfig(
        name="techno",
        type=SubredditType.TOP,
        period=SubredditPeriod.WEEK,
    )

    await get_subreddit_posts(
        mock_client, mock_reddit, subreddit_config, config, {}
    )

    # Verify top was called with time_filter
    mock_subreddit.top.assert_called_once()



"""Testing for the helpers module."""

import os
import tempfile
from datetime import datetime, timedelta
from pathlib import Path
from unittest import mock

import pytest

from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.sync.helpers import (
    parse_sync_command,
    rewrite_track_paths,
    run_sync,
    upload_log,
    webhook,
)

TEST_BUCKET = "s3://some-bucket.com"


@pytest.mark.parametrize("upload", [True, False])
@pytest.mark.parametrize(
    "include_dirs", [[], [Path("path/to/stuff"), Path("path/to/things.mp3")]]
)
@pytest.mark.parametrize(
    "exclude_dirs", [[], [Path("path/to/stuff"), Path("path/to/things.mp3")]]
)
@pytest.mark.parametrize("use_date_modified", [True, False])
@pytest.mark.parametrize("dryrun", [True, False])
def test_parse_sync_command(
    tmpdir,
    config,
    upload,
    include_dirs,
    exclude_dirs,
    use_date_modified,
    dryrun,
):
    """Test for the parse_sync_command function."""
    tmpdir = str(tmpdir)
    setattr(
        config.sync,
        f"{'up' if upload else 'down'}load_include_dirs",
        include_dirs,
    )
    setattr(
        config.sync,
        f"{'up' if upload else 'down'}load_exclude_dirs",
        exclude_dirs,
    )
    config.sync.aws_use_date_modified = not use_date_modified
    config.sync.dryrun = dryrun
    partial_cmd = [
        "aws",
        "s3",
        "sync",
        tmpdir if upload else f"{TEST_BUCKET}/dj/music/",
        f"{TEST_BUCKET}/dj/music/" if upload else tmpdir,
    ]
    cmd = " ".join(parse_sync_command(partial_cmd, config, upload))
    for dir_ in include_dirs:
        assert (
            f"--include {((dir_ / '*').as_posix() if not dir_.suffix else dir_.as_posix())}"
            in cmd
        )
    for dir_ in exclude_dirs:
        assert (
            f"--exclude {((dir_ / '*').as_posix() if not dir_.suffix else dir_.as_posix())}"
            in cmd
        )
    if include_dirs:
        assert "--exclude *" in cmd
    elif exclude_dirs:
        assert "--include *" in cmd
    if use_date_modified:
        assert "--size-only" in cmd
    if dryrun:
        assert "--dryrun" in cmd


def test_rewrite_track_paths(config, rekordbox_xml):
    """Test for the rewrite_track_paths function."""
    user_a_path = Path("/Volumes/first_user_usb/")
    user_b_path = Path("/Volumes/other_user_usb/")
    user_b_xml = rekordbox_xml.parent / "other_user_rekordbox.xml"
    user_b_xml.write_text(
        Path(rekordbox_xml).read_text(encoding="utf-8"), encoding="utf-8"
    )
    config.sync.usb_path = user_a_path

    # Write the second user's usb_path into each track.
    collection = RekordboxCollection(user_b_xml)
    for track in collection.get_tracks().values():
        loc = track.get_location()
        track.set_location(
            loc.parent / str(user_b_path).strip("/") / "DJ Music" / loc.name
        )
    collection.serialize(path=user_b_xml)

    # Replaces all instances of user_b_path in user_b_xml with user_a_path.
    rewrite_track_paths(config, user_b_xml)

    # Assert user_a_path no longer appears and user_b_path is in every track.
    collection = RekordboxCollection(user_b_xml)
    for track in collection.get_tracks().values():
        loc = track.get_location().as_posix()
        assert user_a_path.as_posix() in loc
        assert user_b_path.as_posix() not in loc


@mock.patch("djtools.sync.helpers.Popen")
def test_run_sync(mock_popen, tmpdir):
    """Test for the run_sync function."""
    with open(Path(tmpdir) / "track.mp3", mode="w", encoding="utf-8") as _file:
        _file.write("")
    cmd = ["aws", "s3", "sync", str(tmpdir), f"{TEST_BUCKET}/dj/music/"]
    sync_output = (
        "upload: ../../Volumes/AWEEEEZY/DJ Music/Bass/2022-12-21/track - "
        f"artist.mp3 to {TEST_BUCKET}/dj/music/Bass/2022-12-21/"
        "track - artist.mp3\n"
        "upload: ../../Volumes/AWEEEEZY/DJ Music/Bass/2O22-12-21/other track "
        f"- other artist.mp3 to {TEST_BUCKET}/dj/music/Bass/"
        "2O22-12-21/other track - other artist.mp3\n"
        "upload: ../../Volumes/AWEEEEZY/DJ Music/Techno/2022-12-22/last track "
        f"- last artist.mp3 to {TEST_BUCKET}/dj/music/Techno/"
        "2022-12-22/last track - last artist.mp3"
        "\nirrelevant line"
    )
    # NOTE(a-rich): Windows does not allow opening a temporary file after it's
    # been created. The WAR is to initialize a `NamedTemporaryFile` with the
    # `delete` argument set to `False` and explicitly `.close()` the object
    # before calling `open()` on it.
    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tmp_file:
        tmp_file.write(bytes(sync_output.encode("utf-8")))
        tmp_file.seek(0)
        tmp_file.close()
        with open(tmp_file.name, mode="rb") as tmp_file:
            process = mock_popen.return_value.__enter__.return_value
            process.stdout = tmp_file
            process.wait.return_value = 0
            ret = run_sync(cmd, TEST_BUCKET)
    expected = (
        "Bass/2022-12-21: 1\n\ttrack - artist.mp3\nBass/2O22-12-21: 1\n\tother"
        " track - other artist.mp3\nTechno/2022-12-22: 1\n\tlast track - last "
        "artist.mp3\n"
    )
    assert ret == expected


@mock.patch("djtools.sync.helpers.Popen")
@pytest.mark.parametrize(
    "byte_sequence,expected",
    [
        (
            f"upload: track.mp3 to {TEST_BUCKET}/dj/music/track.mp3\n".encode(
                "utf-8"
            ),
            ".: 1\n\ttrack.mp3\n",
        ),
        (b"\x80", ""),  # This will raise a UnicodeDecodeError.
    ],
)
def test_run_sync_handles_decode_error(
    mock_popen, byte_sequence, expected, tmpdir
):
    """Test for the run_sync function."""
    with open(Path(tmpdir) / "track.mp3", mode="w", encoding="utf-8") as _file:
        _file.write("")
    cmd = ["aws", "s3", "sync", str(tmpdir), f"{TEST_BUCKET}/dj/music/"]
    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tmp_file:
        tmp_file.write(byte_sequence)
        tmp_file.seek(0)
        tmp_file.close()
        with open(tmp_file.name, mode="rb") as tmp_file:
            process = mock_popen.return_value.__enter__.return_value
            process.stdout = tmp_file
            process.wait.return_value = 0
            ret = run_sync(cmd, TEST_BUCKET)
    assert ret == expected


@mock.patch("djtools.sync.helpers.Popen")
def test_run_sync_handles_return_code(mock_popen, tmpdir, caplog):
    """Test for the run_sync function."""
    caplog.set_level("CRITICAL")
    # NOTE(a-rich): subprocess calls to `awscli` need to have unix-style path
    # arguments.
    cmd = [
        "aws",
        "s3",
        "sync",
        Path(tmpdir).as_posix(),
        f"{TEST_BUCKET}/dj/music/",
    ]
    with tempfile.NamedTemporaryFile(mode="wb", delete=False) as tmp_file:
        tmp_file.write(bytes("".encode("utf-8")))
        tmp_file.seek(0)
        tmp_file.close()
        with open(tmp_file.name, mode="rb") as tmp_file:
            process = mock_popen.return_value.__enter__.return_value
            process.stdout = tmp_file
            process.wait.return_value = 1
            msg = (
                f"Failure while syncing: Command '{' '.join(cmd)}' returned "
                f"non-zero exit status {process.wait.return_value}."
            )
            with pytest.raises(Exception, match=msg):
                run_sync(cmd, TEST_BUCKET)
    assert caplog.records[0].message == msg


@mock.patch("djtools.sync.helpers.Popen")
def test_upload_log(mock_popen, tmpdir, config):
    """Test for the upload_log function."""
    config.sync.aws_profile = "DJ"
    now = datetime.now()
    # Windows st_mtime includes fractional seconds which can cause a test
    # failure due to a rounding error.
    one_day_ago = now - timedelta(days=1) - timedelta(seconds=1)
    test_log = f'{now.strftime("%Y-%m-%d")}.log'
    filenames = [
        "__init__.py",
        test_log,
        f'{one_day_ago.strftime("%Y-%m-%d")}.log',
    ]
    ctime = one_day_ago.timestamp()
    for filename in filenames:
        file_path = Path(tmpdir) / filename
        with open(file_path, mode="w", encoding="utf-8") as _file:
            _file.write("")
        if filename != test_log:
            os.utime(file_path, (ctime, ctime))  # pylint: disable=no-member
    process = mock_popen.return_value.__enter__.return_value
    process.wait.return_value = 0
    upload_log(config, Path(tmpdir) / test_log)
    assert len(list(Path(tmpdir).iterdir())) == len(filenames) - 1


def test_upload_log_no_aws_profile(config, caplog):
    """Test for the upload_log function."""
    caplog.set_level("WARNING")
    config.sync.aws_profile = ""
    upload_log(config, "some_file.txt")
    assert (
        caplog.records[0].message == "Logs cannot be backed up without "
        "specifying the config option aws_profile"
    )


@pytest.mark.parametrize(
    "content",
    [
        "*",
        "*" * 2001,
        "*" * 1900 + "\n" + "*" * 99,
        "*" * 1900 + "\n" + "*" * 100,
    ],
)
@mock.patch("requests.post", side_effect=lambda *args, **kwargs: None)
def test_webhook(mock_post, content):
    """Test for the webhook function."""
    url = "https://discord.com/api/webhooks/some-id/"
    content_size_limit = 2000
    posts = len(content) // content_size_limit
    remainder = len(content) % content_size_limit
    posts = posts + 1 if remainder else posts
    webhook(url=url, content_size_limit=content_size_limit, content=content)
    assert mock_post.call_count == max(posts, len(content.split("\n")))


def test_webhook_no_content(caplog):
    """Test for the webhook function."""
    caplog.set_level("INFO")
    url = "https://discord.com/api/webhooks/some-id/"
    content = ""
    webhook(url=url, content=content)
    assert caplog.records[0].message == "There's no content"



"""Testing for the playlists module."""

import pytest

from djtools.collection.base_playlist import Playlist


# pylint:disable=missing-class-docstring,no-method-argument,arguments-differ,protected-access


def test_playlist_raises_type_error():
    """Test Playlist class."""
    with pytest.raises(
        TypeError,
        match="Can't instantiate abstract class Playlist",
    ):
        Playlist()


@pytest.mark.parametrize("enable_aggregation", [True, False])
def test_playlist_adds_aggregate_attribute(enable_aggregation):
    """Test Playlist class."""

    class ConcretePlaylist(Playlist):
        def get_name():
            pass

        def is_folder():
            pass

        def new_playlist():
            pass

        def serialize():
            pass

    playlist = ConcretePlaylist(enable_aggregation=enable_aggregation)
    assert playlist._aggregate == enable_aggregation



"""Update master track Collection with local Collection."""

# pylint: disable=duplicate-code,invalid-name
from argparse import ArgumentParser
from itertools import groupby
from pathlib import Path
from subprocess import Popen

from djtools.configs import build_config
from djtools.collection.platform_registry import PLATFORM_REGISTRY


if __name__ == "__main__":
    # Parse command-line arguments.
    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--config", required=True, help="Path to a config.yaml"
    )
    arg_parser.add_argument(
        "--collection", required=False, help="Path to local collection."
    )
    arg_parser.add_argument(
        "--master-collection",
        required=False,
        help="Path to a remote master collection.",
    )
    arg_parser.add_argument(
        "--master-collection-user",
        required=False,
        default="master",
        help="Username for remote master collection.",
    )
    arg_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite master collection tracks with matching locations.",
    )
    arg_parser.add_argument(
        "--skip-cleanup",
        action="store_true",
        help="Don't remove the temporary master collection after updating its tracks.",
    )
    arg_parser.add_argument(
        "--skip-upload",
        action="store_true",
        help="Don't upload the master collection after updating its tracks.",
    )
    arg_parser.add_argument(
        "--verbose",
        action="store_true",
        help="Display more stuff.",
    )
    args = arg_parser.parse_args()

    # Load config and, if provided, override path to collection.
    config_path = Path(args.config)
    config = build_config(config_path)
    if args.collection:
        collection_path = Path(args.collection)
    else:
        collection_path = config.collection.collection_path

    # Load collection and get a dict of tracks keyed by location.
    collection = PLATFORM_REGISTRY[config.collection.platform]["collection"](
        path=collection_path
    )
    tracks = {
        track.get_location().as_posix(): track
        for _, track in collection.get_tracks().items()
    }

    # Set the path to the remote master collection either from the CLI arg or
    # dynamically. The default path is:
    # <BUCKET-URL>/dj/collections/master/rekordbox_collection
    if args.master_collection:
        remote_master_collection = args.master_collection
    else:
        remote_master_collection = (
            f"{config.sync.bucket_url}/dj/collections/"
            f"{args.master_collection_user}/{config.collection.platform.value}_collection"
        )

    # Download the master collection and get a dict of its tracks too.
    master_collection_path = Path("master_collection.tmp")
    cmd = ["aws", "s3", "cp", remote_master_collection, master_collection_path]
    if collection_path.is_dir():
        cmd.append("--recursive")
    with Popen(cmd) as proc:
        proc.wait()
    master_collection = PLATFORM_REGISTRY[config.collection.platform][
        "collection"
    ](path=master_collection_path)
    master_tracks = master_collection.get_tracks()
    master_tracks = {
        track.get_location().as_posix(): track
        for _, track in master_collection.get_tracks().items()
    }

    # The old master collection tracks dict is needed for verbose info.
    old_master_tracks = dict(master_tracks)

    # If not overwriting master collection tracks, then filter the new tracks
    # by those that don't have a key in the master collection tracks dict.
    tracks = (
        tracks
        if args.overwrite
        else {k: v for k, v in tracks.items() if k not in master_tracks}
    )

    # Update master collection tracks
    master_tracks = {**master_tracks, **tracks}

    # Index only new tracks for downstream processing.
    new_tracks = [
        value for key, value in tracks.items() if key not in old_master_tracks
    ]

    # Print track counts for the different collections as well as the new
    # tracks. Print the new tracks grouped by date added.
    if args.verbose:
        print(
            f"Local tracks: {len(tracks)}\n"
            f"Master tracks: {len(old_master_tracks)}\n"
            f"New tracks: {len(new_tracks)}"
        )
        for date, tracks in groupby(
            sorted(new_tracks, key=lambda x: x.get_date_added()),
            key=lambda x: x.get_date_added(),
        ):
            tracks = list(tracks)
            print(f"{date.strftime('%Y-%m-%d')}: {len(tracks)}")
            for track in tracks:
                print(f"\t{track.get_location().name}")

    # Don't waste time serializing or uploading if there aren't any new tracks.
    if new_tracks or args.overwrite:
        master_collection.set_tracks(master_tracks)
        master_collection.serialize()
    elif not args.overwrite:
        args.skip_upload = True

    # Upload the newly updated master collection.
    if not args.skip_upload:
        cmd = [
            "aws",
            "s3",
            "cp",
            master_collection_path,
            remote_master_collection,
        ]
        if collection_path.is_dir():
            cmd.append("--recursive")
        with Popen(cmd) as proc:
            proc.wait()

    # Clean up the temporary master collection.
    if not args.skip_cleanup:
        master_collection_path.unlink()



"""Testing for the playlists module."""

import pytest

from djtools.collection.rekordbox_playlist import RekordboxPlaylist


def test_rekordboxplaylist_getitem(rekordbox_playlist):
    """Test RekordboxPlaylist class."""
    try:
        _ = rekordbox_playlist[0]
    except KeyError:
        assert False, "RekordboxPlaylist.__getitem__ failed!"


@pytest.mark.parametrize("playlist,expected", [(None, 3), ("Hip Hop", 1)])
def test_rekordboxplaylist_len(playlist, expected, rekordbox_collection):
    """Test RekordboxPlaylist class."""
    # Get a playlist containing tracks (len returns number of tracks).
    if playlist:
        playlist = rekordbox_collection.get_playlists(playlist)[0]
    # Get the root playlist (len returns number of playlists at the root).
    else:
        playlist = rekordbox_collection.get_playlists()
    assert len(playlist) == expected


@pytest.mark.parametrize("index", [None, 0])
def test_rekordboxplaylist_add_playlist(index, rekordbox_playlist):
    """Test RekordboxPlaylist class."""
    test_playlist_name = "TEST"
    test_playlist = RekordboxPlaylist.new_playlist(
        test_playlist_name, tracks={}
    )
    assert not rekordbox_playlist.get_playlists(test_playlist_name)
    rekordbox_playlist.add_playlist(test_playlist, index=index)
    if index is not None:
        assert rekordbox_playlist[index] == test_playlist
    else:
        assert rekordbox_playlist[-1] == test_playlist
    rekordbox_playlist.remove_playlist(test_playlist)


def test_rekordboxplaylist_add_playlist_raises_runtimeerror_when_appending_to_non_folder(
    rekordbox_playlist,
):
    """Test RekordboxPlaylist class."""
    leaf_playlist = rekordbox_playlist.get_playlists("Dark")[0]
    with pytest.raises(
        RuntimeError,
        match="You can't append to a non-folder Playlist",
    ):
        leaf_playlist.add_playlist("")


@pytest.mark.parametrize("playlist_name", ["Hip Hop", "Dark"])
def test_rekordboxplaylist_get_name_and_get_playlists(
    playlist_name, rekordbox_playlist
):
    """Test RekordboxPlaylist class."""
    for playlist in rekordbox_playlist.get_playlists(playlist_name):
        assert playlist.get_name() == playlist_name


def test_rekordboxplaylist_get_parent(rekordbox_collection):
    """Test RekordboxPlaylist class."""
    root_playlist = rekordbox_collection.get_playlists()
    assert not root_playlist.get_parent()
    for playlist in root_playlist:
        assert playlist.get_parent() == root_playlist


def test_rekordboxplaylist_get_playlists_raises_runtimeerror_with_non_folder(
    rekordbox_playlist,
):
    """Test RekordboxPlaylist class."""
    playlist = rekordbox_playlist.get_playlists("Hip Hop")[0]
    with pytest.raises(
        RuntimeError,
        match=(
            f'Playlist "{playlist.get_name()}" is not a folder so you cannot '
            f"call get_playlists on it."
        ),
    ):
        playlist.get_playlists()


def test_rekordboxplaylist_get_playlists_with_glob(rekordbox_playlist):
    """Test RekordboxPlaylist class."""
    for playlist in rekordbox_playlist.get_playlists("Hip*", glob=True):
        assert "Hip" in playlist.get_name()


def test_rekordboxplaylist_get_tracks(rekordbox_track):
    """Test RekordboxPlaylist class."""
    tracks = {rekordbox_track.get_id(): rekordbox_track}
    playlist = RekordboxPlaylist.new_playlist("TEST", tracks=tracks)
    assert playlist.get_tracks() == tracks


@pytest.mark.parametrize(
    "playlist,expected", [("Genres", True), ("Hip Hop", False)]
)
def test_rekordboxplaylist_is_folder(playlist, expected, rekordbox_playlist):
    """Test RekordboxPlaylist class."""
    assert (
        rekordbox_playlist.get_playlists(playlist)[0].is_folder() == expected
    )


@pytest.mark.parametrize(
    "playlist_name,tracks,playlists,is_folder",
    [("TEST", {}, None, False), ("something", None, [], True)],
)
def test_rekordboxplaylist_new_playlist(
    playlist_name, tracks, playlists, is_folder
):
    """Test RekordboxPlaylist class."""
    playlist = RekordboxPlaylist.new_playlist(
        playlist_name, tracks=tracks, playlists=playlists
    )
    assert playlist.get_name() == playlist_name
    assert playlist.is_folder() == is_folder


def test_rekordboxplaylist_newplaylist_raises_runtimeerror_with_both_playlists_and_tracks():
    """Test RekordboxPlaylist class."""
    with pytest.raises(
        RuntimeError,
        match=(
            "You must not provide both a list of RekordboxPlaylists and a "
            "list of RekordboxTracks"
        ),
    ):
        RekordboxPlaylist.new_playlist("", playlists=[], tracks={})


def test_rekordboxplaylist_newplaylist_raises_runtimeerror_with_no_playlists_or_tracks():
    """Test RekordboxPlaylist class."""
    with pytest.raises(
        RuntimeError,
        match=(
            "You must provide either a list of RekordboxPlaylists or a "
            "list of RekordboxTracks"
        ),
    ):
        RekordboxPlaylist.new_playlist("", playlists=None, tracks=None)


def test_rekordboxplaylist_remove_playlists(rekordbox_playlist):
    """Test RekordboxPlaylist class."""
    test_playlist_name = "TEST"
    test_playlist = RekordboxPlaylist.new_playlist(
        test_playlist_name, tracks={}
    )
    rekordbox_playlist.add_playlist(test_playlist)
    assert rekordbox_playlist.get_playlists(test_playlist_name)
    rekordbox_playlist.remove_playlist(test_playlist)
    assert not rekordbox_playlist.get_playlists(test_playlist_name)


def test_rekordboxplaylist_remove_playlists_raises_runtimeerror_when_removing_folder(
    rekordbox_playlist,
):
    """Test RekordboxPlaylist class."""
    playlist = rekordbox_playlist.get_playlists("Hip Hop")[0]
    with pytest.raises(
        RuntimeError,
        match="Can't remove playlist from a non-folder playlist.",
    ):
        playlist.remove_playlist("")


def test_rekordboxplaylist_serialization(
    rekordbox_playlist_tag, rekordbox_track
):
    """Test RekordboxPlaylist class."""
    playlist = RekordboxPlaylist(
        rekordbox_playlist_tag, tracks={"2": rekordbox_track}
    )
    assert (
        repr(playlist)
        == """RekordboxPlaylist(Name="ROOT", Type="0", Count="2",
    playlists=[
        RekordboxPlaylist(Name="Genres", Type="0", Count="1",
            playlists=[
                RekordboxPlaylist(Name="Hip Hop", Type="1", Entries="1")
            ]
        ),
        RekordboxPlaylist(Name="My Tags", Type="0", Count="1",
            playlists=[
                RekordboxPlaylist(Name="Dark", Type="1", Entries="0")
            ]
        )
    ]
)"""
    )
    assert str(playlist) == str(rekordbox_playlist_tag)
    assert playlist.serialize() == rekordbox_playlist_tag


def test_rekordboxplaylist_set_parent():
    """Test RekordboxPlaylist class."""
    child_playlist = RekordboxPlaylist.new_playlist("Child", tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        "Parent", playlists=[child_playlist]
    )
    assert child_playlist.get_parent() is None
    child_playlist.set_parent(parent_playlist)
    assert child_playlist.get_parent() is parent_playlist


def test_rekordboxplaylist_set_tracks(rekordbox_track):
    """Test RekordboxPlaylist class."""
    old_tracks = {}
    playlist = RekordboxPlaylist.new_playlist("Test", tracks=old_tracks)
    assert playlist.get_tracks() == old_tracks
    new_tracks = {rekordbox_track.get_id(): rekordbox_track}
    playlist.set_tracks(new_tracks)
    assert playlist.get_tracks() == new_tracks



"""Enumerations for the spotify package.

This module contains enum types used by the spotify config and helpers
to avoid cyclic imports.
"""

from enum import Enum

import yaml


class SubredditPeriod(Enum):
    """Time period for subreddit queries."""

    ALL = "all"
    DAY = "day"
    HOUR = "hour"
    MONTH = "month"
    WEEK = "week"
    YEAR = "year"


def subreddit_period_representer(dumper, data):
    """YAML representer for SubredditPeriod."""
    return dumper.represent_scalar("!SubredditPeriod", data.value)


def subreddit_period_constructor(loader, node):
    """YAML constructor for SubredditPeriod."""
    return SubredditPeriod(loader.construct_scalar(node))


yaml.add_representer(SubredditPeriod, subreddit_period_representer)
yaml.add_constructor("!SubredditPeriod", subreddit_period_constructor)


class SubredditType(Enum):
    """Type of subreddit sort."""

    CONTROVERSIAL = "controversial"
    HOT = "hot"
    NEW = "new"
    RISING = "rising"
    TOP = "top"


def subreddit_type_representer(dumper, data):
    """YAML representer for SubredditType."""
    return dumper.represent_scalar("!SubredditType", data.value)


def subreddit_type_constructor(loader, node):
    """YAML constructor for SubredditType."""
    return SubredditType(loader.construct_scalar(node))


yaml.add_representer(SubredditType, subreddit_type_representer)
yaml.add_constructor("!SubredditType", subreddit_type_constructor)



"""This module contains helper functions used by the spotify module.

Most of the heavy lifting is now delegated to the spotify-tools library.
This module provides DJ-Tools specific wrappers and configuration handling.
"""

import logging
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

import asyncpraw as praw
import yaml

from spotify_tools import (
    Client,
    SpotifyConfig,
    filter_tracks_by_similarity,
    is_duplicate_track,
    search_track_fuzzy,
)


logger = logging.getLogger(__name__)
BaseConfig = Type["BaseConfig"]
DJToolsSpotifyConfig = Type["SpotifyConfig"]
SubredditConfig = Type["SubredditConfig"]


def get_playlist_ids() -> Dict[str, str]:
    """Load Spotify playlist names -> IDs lookup.

    Returns:
        Dictionary of Spotify playlist names mapped to playlist IDs.
    """
    playlist_ids = {}
    ids_path = (
        Path(__file__).parent.parent / "configs" / "spotify_playlists.yaml"
    )
    if ids_path.exists():
        with open(ids_path, mode="r", encoding="utf-8") as _file:
            playlist_ids = (
                yaml.load(_file, Loader=yaml.FullLoader) or playlist_ids
            )

    return playlist_ids


def get_reddit_client(config: BaseConfig) -> praw.Reddit:
    """Instantiate a Reddit API client.

    Args:
        config: Configuration object.

    Returns:
        Reddit API client.
    """
    reddit = praw.Reddit(
        client_id=config.spotify.reddit_client_id,
        client_secret=config.spotify.reddit_client_secret,
        user_agent=config.spotify.reddit_user_agent,
        timeout=30,
    )

    return reddit


def get_spotify_client(
    config: Union[BaseConfig, DJToolsSpotifyConfig],
) -> Client:
    """Instantiate a Spotify API client using spotify-tools.

    Args:
        config: Configuration object.

    Returns:
        Spotify API client.
    """
    try:
        spotify_config = getattr(config, "spotify")
    except AttributeError:
        spotify_config = config

    st_config = SpotifyConfig(
        CLIENT_ID=spotify_config.spotify_client_id,
        CLIENT_SECRET=spotify_config.spotify_client_secret,
        REDIRECT_URI=spotify_config.spotify_redirect_uri,
    )

    return Client(
        config=st_config,
        scopes=["playlist-modify-public"],
        cache_path=Path(__file__).parent / ".spotify.cache",
    )


def filter_results(
    spotify: Client,
    results: Dict,
    threshold: float,
    title: str,
    artist: str,
) -> Tuple[Dict[str, Any], float]:
    """Filter Spotify search results to find best matching track.

    Args:
        spotify: Spotify client.
        results: Spotify search results.
        threshold: Minimum Levenshtein distance.
        title: Potential title of a track.
        artist: Potential artist of a track.

    Returns:
        Tuple of track object (as dict) and similarity score.
    """
    # Convert spotify-tools Track objects to dicts for backward compatibility
    tracks = results.get("tracks", {}).get("items", [])

    # Use spotify-tools filtering
    from spotify_tools.schemas import Track

    track_objects = [Track.model_validate(t) for t in tracks if t]

    matches = filter_tracks_by_similarity(
        track_objects, title, artist, threshold
    )

    # Paginate through remaining results
    while results.get("tracks", {}).get("next"):
        try:
            results = spotify.next(results["tracks"])
            if not results:
                break
            tracks = results.get("tracks", {}).get("items", [])
            if not tracks:
                tracks = results.get("items", [])
            track_objects = [Track.model_validate(t) for t in tracks if t]
            matches.extend(
                filter_tracks_by_similarity(
                    track_objects, title, artist, threshold
                )
            )
        except Exception:
            logger.warning(f"Failed to get next tracks for {title, artist}")
            break

    if not matches:
        return {}, 0.0

    best = max(matches, key=lambda m: m.score)
    # Convert back to dict format for backward compatibility
    return best.track.model_dump(), best.score


async def get_subreddit_posts(
    spotify: Client,
    reddit: praw.Reddit,
    subreddit: SubredditConfig,
    config: BaseConfig,
    praw_cache: Dict[str, bool],
) -> Tuple[List[Tuple[str]], Dict[str, Union[str, int]]]:
    """Filters subreddit submissions and resolves them to Spotify tracks.

    Args:
        spotify: Spotify client.
        reddit: Reddit client.
        subreddit: SubredditConfig object.
        config: Configuration object.
        praw_cache: Cached praw submissions.

    Returns:
        List of Spotify track ("id", "name") tuples and SubredditConfig.
    """
    from concurrent.futures import as_completed, ThreadPoolExecutor
    from tqdm import tqdm

    sub = await reddit.subreddit(subreddit.name)
    func = getattr(sub, subreddit.type.value)
    kwargs = {"limit": config.spotify.spotify_playlist_post_limit}
    from djtools.spotify.enums import SubredditType

    if subreddit.type == SubredditType.TOP:
        kwargs["time_filter"] = subreddit.period

    subs = [
        x
        async for x in _catch(
            func(**kwargs), message="Failed to retrieve Reddit submission"
        )
    ]

    msg = f'Filtering {len(subs)} "r/{subreddit.name}" {subreddit.type.value} posts'
    logger.info(msg)

    submissions = []
    for submission in tqdm(subs, desc=msg):
        if submission.id in praw_cache:
            continue
        submissions.append(submission)
        praw_cache[submission.id] = True

    new_tracks = []
    if submissions:
        msg = (
            f"Searching Spotify for {len(submissions)} new submission(s) from "
            f'"r/{subreddit.name}"'
        )
        logger.info(msg)
        payload = zip(
            submissions,
            [spotify] * len(submissions),
            [config.spotify.spotify_playlist_fuzz_ratio] * len(submissions),
        )

        with ThreadPoolExecutor(max_workers=8) as executor:
            futures = [executor.submit(_process, *args) for args in payload]

            with tqdm(total=len(futures), desc=msg) as pbar:
                new_tracks = []
                for future in as_completed(futures):
                    new_tracks.append(future.result())
                    pbar.update(1)

        new_tracks = [track for track in new_tracks if track]
        logger.info(
            f"Got {len(new_tracks)} Spotify track(s) from new "
            f'"r/{subreddit.name}" posts'
        )
    else:
        logger.info(f'No new submissions from "r/{subreddit.name}"')

    return new_tracks, subreddit


def populate_playlist(
    playlist_name: str,
    playlist_ids: Dict[str, str],
    spotify_username: str,
    spotify: Client,
    tracks: List[Tuple[str]],
    playlist_limit: Optional[int] = None,
    verbosity: Optional[int] = 0,
) -> Dict[str, str]:
    """Inserts tracks into either a new playlist or an existing one.

    Args:
        playlist_name: Name of the playlist.
        playlist_ids: Lookup of playlist IDs.
        spotify_username: Spotify user's username.
        spotify: Spotify client.
        tracks: List of (track_id, track_name) tuples.
        playlist_limit: Maximum number of tracks allowed in a playlist.
        verbosity: Logging verbosity level.

    Returns:
        Updated playlist IDs.
    """
    playlist_id = playlist_ids.get(playlist_name)
    playlist = None

    if playlist_id and tracks:
        playlist = _update_existing_playlist(
            spotify,
            playlist_id,
            tracks,
            playlist_limit,
            verbosity,
        )
    elif tracks:
        logger.warning(
            f"Unable to get ID for {playlist_name}...creating a new playlist"
        )
        playlist = _build_new_playlist(
            spotify, spotify_username, playlist_name, tracks
        )
        playlist_ids[playlist_name] = playlist["id"]
    elif playlist_id:
        playlist = spotify.playlist(playlist_id)

    if playlist:
        logger.info(
            f'"{playlist["name"]}": '
            f'{playlist["external_urls"].get("spotify")}'
        )

    return playlist_ids


def write_playlist_ids(playlist_ids: Dict[str, str]):
    """Write playlist IDs to file.

    Args:
        playlist_ids: Dictionary of Spotify playlist names mapped to playlist
            IDs.
    """
    ids_path = (
        Path(__file__).parent.parent / "configs" / "spotify_playlists.yaml"
    )
    with open(ids_path, mode="w", encoding="utf-8") as _file:
        yaml.dump(playlist_ids, _file)


def _build_new_playlist(
    spotify: Client,
    username: str,
    playlist_name: str,
    new_tracks: List[Tuple[str]],
) -> Dict[str, Any]:
    """Creates a new playlist from a list of track IDs / URLs.

    Args:
        spotify: Spotify client.
        username: Spotify username.
        playlist_name: Name for the new playlist.
        new_tracks: List of (track_id, track_name) tuples.

    Returns:
        Playlist object for the newly constructed playlist.
    """
    ids = [t[0] for t in new_tracks]
    playlist = spotify.user_playlist_create(
        username, name=playlist_name.title()
    )
    spotify.playlist_add_items(playlist["id"], ids)

    return playlist


async def _catch(
    generator: AsyncGenerator, message: Optional[str] = ""
) -> Any:
    """Permits one-line try/except logic for async comprehensions.

    Args:
        generator: Async generator.
        message: Prefix message for logger warning.

    Yields:
        Items from the AsyncGenerator.
    """
    while True:
        try:
            yield await anext(generator)
        except StopAsyncIteration:
            return
        except Exception as exc:
            logger.warning(f"{message}: {exc}" if message else exc)
            continue


def _filter_tracks(
    tracks: List[Dict], threshold: float, title: str, artist: str
) -> List[Tuple[Dict[str, Any], float]]:
    """Filter tracks by Levenshtein distance on artist and name fields.

    Args:
        tracks: Spotify search results.
        threshold: Minimum Levenshtein distance.
        title: Potential title of a track.
        artist: Potential artist of a track.

    Returns:
        List of (track_dict, similarity_score) tuples.
    """
    from spotify_tools.schemas import Track

    track_objects = [Track.model_validate(t) for t in tracks if t]
    matches = filter_tracks_by_similarity(
        track_objects, title, artist, threshold
    )
    return [(m.track.model_dump(), m.score) for m in matches]


def _fuzzy_match(
    spotify: Client, title: str, threshold: float
) -> Optional[Tuple[str, str]]:
    """Attempts to find a Spotify track matching the title.

    Args:
        spotify: Spotify client.
        title: Submission title.
        threshold: Minimum Levenshtein distance.

    Returns:
        Tuple of (track_id, "track_name - artists") or None.
    """
    parts = _parse_title(title)
    if not all(parts):
        return None

    # Try both orderings (title-artist and artist-title)
    for track, artist in [parts, parts[::-1]]:
        try:
            result = search_track_fuzzy(
                spotify, track, artist, threshold=threshold, limit=50
            )
            if result and result.track:
                track_obj = result.track
                artists = ", ".join(
                    a.name for a in (track_obj.artists or []) if a.name
                )
                return (track_obj.id, f"{track_obj.name} - {artists}")
        except Exception as exc:
            logger.error(f'Error searching for "{track} - {artist}": {exc}')
            continue

    return None


def _parse_title(title: str) -> List[str]:
    """Attempts to split submission title into (track name, artist(s)).

    Args:
        title: Submission title.

    Returns:
        Pair of strings: [title, artist] or [None, None] if unparseable.
    """
    try:
        title_part, artist = map(str.strip, title.split(" - "))
    except ValueError:
        try:
            title_part, artist = map(str.strip, title.lower().split(" by "))
        except ValueError:
            return [None, None]

    # Remove parenthetical and bracketed content
    title_part = title_part.split("(")[0].strip()
    title_part = title_part.split("[")[0].strip()
    artist = artist.split("(")[0].strip()
    artist = artist.split("[")[0].strip()

    return [title_part, artist]


def _process(
    submission: praw.models.Submission,
    spotify: Client,
    threshold: float,
) -> Optional[Tuple[str, str]]:
    """Worker thread process for resolving a submission to a track.

    Args:
        submission: Reddit Submission object.
        spotify: Spotify API client.
        threshold: Minimum Levenshtein distance.

    Returns:
        Tuple of (track_id/url, title) or None.
    """
    # Direct Spotify links can be returned as-is
    if "spotify.com/track/" in submission.url:
        return (submission.url, submission.title)

    return _fuzzy_match(spotify, submission.title, threshold)


def _track_name_too_similar(
    track: str,
    playlist_track_names: set,
) -> bool:
    """Check if a track is too similar to existing tracks.

    Args:
        track: Track name to check.
        playlist_track_names: Set of existing track names.

    Returns:
        True if the track is a duplicate.
    """
    if is_duplicate_track(track, playlist_track_names, threshold=90.0):
        for other in playlist_track_names:
            from fuzzywuzzy import fuzz

            if fuzz.ratio(track.lower(), other.lower()) > 90:
                logger.warning(
                    f'Candidate new track "{track}" is too similar to '
                    f'existing track "{other}"'
                )
                return True
    return False


def _update_existing_playlist(
    spotify: Client,
    playlist_id: str,
    new_tracks: List[Tuple[str, str]],
    limit: Optional[int],
    verbosity: int,
) -> Dict[str, Any]:
    """Updates an existing playlist with new tracks (LIFO queue behavior).

    Args:
        spotify: Spotify client.
        playlist_id: Spotify playlist ID.
        new_tracks: List of (track_id, track_name) tuples.
        limit: Maximum number of tracks in playlist.
        verbosity: Verbosity level.

    Returns:
        Playlist object.
    """
    import sys

    if limit is None:
        limit = sys.maxsize  # pragma: no cover

    # Get current playlist and tracks
    playlist_object = spotify.playlist(playlist_id)
    _playlist = playlist_object
    tracks = _playlist["tracks"]["items"]

    # Paginate to get all tracks
    try:
        while _playlist["tracks"]["next"]:
            _playlist = spotify.next(_playlist["tracks"])
            try:
                tracks.extend(_playlist["tracks"]["items"])
            except KeyError:
                tracks.extend(_playlist["items"])
    except KeyError:
        tracks.extend(_playlist.get("items", []))
    except Exception as exc:
        logger.error(f"Failed to get tracks from playlist: {exc}")

    # Build sets of existing track IDs and names
    track_count = len(tracks)
    track_index = 0
    add_payload = []
    tracks_added = []
    remove_payload = []
    tracks_removed = []
    ids = set()
    playlist_track_names = set()

    for track in tracks:
        track_data = track["track"]
        ids.add(track_data["id"])
        artists = ", ".join([x["name"] for x in track_data["artists"]])
        playlist_track_names.add(f'{track_data["name"]} - {artists}')

    # Process new tracks
    for id_, track_name in new_tracks:
        # Resolve URL to track ID if needed
        if "spotify.com/track/" in id_:
            resp = spotify.track(id_)
            id_ = resp["id"]
            artists = ", ".join([x["name"] for x in resp["artists"]])
            track_name = f'{resp["name"]} - {artists}'

        if id_ in ids:
            logger.warning(
                f'Candidate new track "{track_name}" is already in the playlist'
            )
            continue

        if _track_name_too_similar(track_name, playlist_track_names):
            continue  # pragma: no cover

        tracks_added.append(track_name)
        add_payload.append(id_)

        # Remove oldest track if we'd exceed limit
        if track_count + len(tracks_added) > limit:
            _track = tracks.pop(0)["track"]
            artists = ", ".join([x["name"] for x in _track["artists"]])
            tracks_removed.append(f'{_track["name"]} - {artists}')
            remove_payload.append(
                {"uri": _track["uri"], "positions": [track_index]}
            )
            track_index += 1
            track_count -= 1

    # Log what we're doing
    if tracks_added:
        logger.info(f"{len(tracks_added)} new tracks added")
        if verbosity > 0:
            for track in tracks_added:
                logger.info(f"\t{track}")

    if tracks_removed:
        logger.info(f"{len(tracks_removed)} old tracks removed")
        if verbosity > 0:  # pragma: no cover
            for track in tracks_removed:
                logger.info(f"\t{track}")

    if not (tracks_added or tracks_removed):
        logger.info("No tracks added or removed")

    # Execute the changes
    if remove_payload:
        spotify.playlist_remove_specific_occurrences_of_items(
            playlist_id, remove_payload
        )
    if add_payload:
        spotify.playlist_add_items(playlist_id, add_payload)

    return playlist_object



"""This script is used to create a new playlist comprised of tracks which do
not appear in a given playlist (or, if using '--folder', any playlists in a
folder).
"""

from argparse import ArgumentParser
import logging
from traceback import format_exc

from bs4 import BeautifulSoup
from bs4.element import Tag


logger = logging.getLogger(__name__)


def create_playlist(_soup, tracks, new_playlist_name):
    """Creates a new playlist with name 'new_playlist_name' and inserts all the
    dangling tracks into it.

    Args:
        _soup (bs4.BeautifulSoup): parsed XML
        tracks (dict): map of TrackIDs to XML tags for dangling tracks
        new_playlist (str): new playlist name
    """
    playlists_root = _soup.find_all("NODE", {"Name": "ROOT", "Type": "0"})[0]
    new_playlist = _soup.new_tag(
        "NODE",
        Name=new_playlist_name,
        Type="1",
        KeyType="0",
        Entries=str(len(tracks)),
    )
    playlists_root.insert(0, new_playlist)
    for track_id, _ in tracks.items():
        new_playlist.append(_soup.new_tag("TRACK", Key=track_id))


def get_tracks(_soup, name, folder):
    """Finds a playlist (or folder) matching 'name' and identifies tracks which
    do not belong to the given playlist (or any of the playlists within the
    folder).

    Args:
        _soup (bs4.BeautifulSoup): parsed XML
        name (str): playlist / folder name used to exclude tracks
        folder (bool): whether or not 'name' is a folder

    Raises:
        ValueError: XML must contain playlist (or folder) matching 'name'

    Returns:
        dict: map of TrackIDs to XML tags for dangling tracks
    """
    tracks = {
        x["TrackID"]: x for x in _soup.find_all("TRACK") if x.get("Location")
    }
    try:
        node = _soup.find_all(
            "NODE", {"Name": name, "Type": "0" if folder else "1"}
        )[0]
    except IndexError:
        msg = f'Failed to find {"folder" if folder else "playlist"} {name}'
        logger.critical(msg)
        raise ValueError(msg) from IndexError

    if folder:
        playlists = node.find_all("NODE", {"Type": "1"})
    else:
        playlists = [node]

    logger.info(f"Tracks: {len(tracks)}")
    logger.info(f"Playlists: {len(playlists)}")

    for playlist in playlists:
        for track in playlist.children:
            if not (isinstance(track, Tag) and tracks.get(track["Key"])):
                continue
            del tracks[track["Key"]]

    logger.info(f"Dangling tracks: {len(tracks)}")

    return tracks


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--xml_path", required=True, type=str, help="path to Rekordbox XML"
    )
    parser.add_argument(
        "--name",
        type=str,
        required=True,
        help="playlist / folder name used to find missing tracks",
    )
    parser.add_argument(
        "--folder",
        action="store_true",
        help='set if "--name" is folder instead of playlist',
    )
    parser.add_argument(
        "--new_playlist",
        type=str,
        default="Dangling Tracks",
        help="name of playlist to create with dangling tracks",
    )
    args = parser.parse_args()

    with open(args.xml_path, encoding="utf-8") as _file:
        soup = BeautifulSoup(_file.read(), "xml")
    try:
        create_playlist(
            soup, get_tracks(soup, args.name, args.folder), args.new_playlist
        )
    except ValueError as exc:
        logger.critical(exc)
    except Exception:
        logger.critical(f"Some error occured: {format_exc()}")

    with open(args.xml_path, mode="wb", encoding=soup.orignal_encoding) as f:
        f.write(soup.prettify("utf-8"))



"""This module is a script for modifying a Collection.

More specifically, it queries Spotify for each track and attempts to resolve
the year, album, and label. In addition, it removes "situation" tags
{"Opener", "Build", "Peak Time"} from the Comments field. This is done because
I'm now re-purposing the "Rating" field to convey the energy level of tracks.
"""

# pylint: disable=redefined-outer-name,duplicate-code
from argparse import ArgumentParser
from concurrent.futures import as_completed, ThreadPoolExecutor
import os
from pathlib import Path
import re
import sys

from tqdm import tqdm

from djtools.configs.helpers import build_config
from djtools.collection.platform_registry import PLATFORM_REGISTRY


def remove_tags_thread(track, tag_regex, remove_tags):
    """Threaded function.

    If the track has "My Tag" data in the Comments field, remove the
    provided tags.

    Args:
        track: A Track object.
        tag_regex: Regular expression to match tags in the comments.
        remove_tags: List of tags to exclude from the comments.
    """
    field = getattr(track, "_Comments")
    tags = re.search(tag_regex, field)
    if not tags:
        return
    comment_prefix = field.split("/* ")[0].strip()
    comment_suffix = field.split(" */")[-1].strip()
    new_tags = [
        tag.strip()
        for tag in tags.group().split("/")
        if tag.strip() not in remove_tags
    ]
    new_tags = " / ".join(new_tags)
    if new_tags:
        setattr(
            track,
            "_Comments",
            f"{comment_prefix} /* {new_tags} */ {comment_suffix}",
        )
    else:
        setattr(track, "_Comments", f"{comment_prefix} {comment_suffix}")


if __name__ == "__main__":
    arg_parser = ArgumentParser()
    arg_parser.add_argument("--config", help="Path to a config.yaml")
    arg_parser.add_argument(
        "--output_collection", help="Path to output collection"
    )
    arg_parser.add_argument(
        "--remove-tags", nargs="+", help="List of tags to remove"
    )
    args = arg_parser.parse_args()

    # Build config, instantiate collection, and get tracks.
    config_path = Path(args.config)
    config = build_config(config_path)
    collection = PLATFORM_REGISTRY[config.platform]["collection"](
        path=config.collection_path
    )
    tracks = collection.get_tracks().values()

    if not args.remove_tags:
        sys.exit("You must provide --remove-tags!")

    # Remove tags from a particular field.
    tag_regex = re.compile(r"(?<=\/\*).*(?=\*\/)")
    remove_tags = set(args.remove_tags)
    with (
        tqdm(
            total=len(tracks),
            desc=f"Removing tags from {args.remove_tags_fields}",
        ) as pbar,
        ThreadPoolExecutor(max_workers=os.cpu_count() * 4) as pool,
    ):
        futures = [
            pool.submit(remove_tags_thread, track, tag_regex, remove_tags)
            for track in tracks
        ]
        for future in as_completed(futures):
            future.result()
            pbar.update(1)

    playlist_class = PLATFORM_REGISTRY[config.platform]["playlist"]
    playlist = playlist_class.new_playlist(
        name=args.playlist_name, tracks=tracks
    )
    collection.add_playlist(playlist)
    output = args.output_collection if args.output_collection else config_path
    collection.serialize(output)
